import logging
from abc import ABC, abstractmethod
from enum import Enum
from typing import Optional, Union, Sequence
from typing_extensions import Protocol

import torch

from ..lstnet.lstnet_modules import LSTNetwork
from ..mlp.mlp_modules import MultiLayerPerceptron
from ... import ActivationFunction
from ....util.string import object_repr, ToStringMixin

log = logging.getLogger(__name__)


class EncoderProtocol(Protocol):
    def forward(self, x: torch.Tensor, lengths: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        :param x: a tensor of shape (batch_size, seq_length=max(lengths), history_features) containing the sequence of
            history features to encode
        :param lengths: an optional tensor of shape (batch_size) containing the lengths of the sequences in `x`
        :return: a tensor of shape (batch_size, latent_dim) containing the encodings
        """
        pass


class DecoderProtocol(Protocol):
    def forward(self,
            latent: torch.Tensor,
            target_features: Optional[torch.Tensor] = None,
            target_lengths: Optional[torch.Tensor] = None) -> torch.Tensor:
        """
        :param latent: a tensor of shape (batch_size, latent_dim) containing the latent representations
        :param target_features: a tensor of shape (batch_size, target_seq_length=max(target_lengths), target_feature_dim)
        :param target_lengths: a tensor of shape (batch_size) containing the lengths of sequences in `target_features`
        :return: a tensor of shape (batch_size, output_dim) or (batch_size, target_seq_length, output_dim) containing the predictions,
            where the shape depends on the use case and can vary depending on the needs
        """
        pass


class PredictorProtocol(Protocol):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        :param x: a tensor of shape (batch_size, input_dim) an intermediate representation
        :return: a tensor of shape (batch_size, output_dim)
        """
        pass


# TODO: Should use intersection type A & B once we switch to Python 3.9+
TDecoder = Union[DecoderProtocol, torch.nn.Module]
TEncoder = Union[EncoderProtocol, torch.nn.Module]
TPredictor = Union[PredictorProtocol, torch.nn.Module]


class EncoderFactory(ToStringMixin, ABC):
    """
    Represents a factory for an encoder modules that map a sequence of items to a latent vector
    """

    @abstractmethod
    def create_encoder(self, input_dim: int, latent_dim: int) -> TEncoder:
        """
        :param input_dim: the input dimension per sequence item
        :param latent_dim: the latent vector dimension that is to be generated by the encoder
        :return: a torch module satisfying :class:`EncoderProtocol`
        """
        pass


class DecoderFactory(ToStringMixin, ABC):
    @abstractmethod
    def create_decoder(self, latent_dim: int, target_feature_dim: int) -> TDecoder:
        """
        :param latent_dim: the latent vector size which is used for the representation of the history
        :param target_feature_dim: the number of dimensions/features that are given for each prediction to be made
            (each future sequence item)
        :return: a torch module satisfying :class:`DecoderProtocol`
        """
        pass


class PredictorFactory(ToStringMixin, ABC):
    """
    Represents a factory for predictor components which sample map from an intermediate representation to the
    desired output dimension.
    """
    def create_predictor(self, input_dim: int, output_dim: int) -> TPredictor:
        """
        :param input_dim: the input dimension
        :param output_dim: the output dimension
        :return: a module which maps an input with dimension `input_dim` to the desired prediction dimension (`output_dim`)
        """
        pass


class LinearPredictorFactory(PredictorFactory):
    """A factory for predictors consisting only of a linear layer (without subsequent activation)"""
    def create_predictor(self, input_dim: int, output_dim: int) -> torch.nn.Module:
        return torch.nn.Linear(input_dim, output_dim)


class MLPPredictorFactory(PredictorFactory):
    """A factor for predictors that are multi-layer perceptrons"""
    def __init__(self,
            hidden_dims: Sequence[int] = (),
            hid_activation_fn: ActivationFunction = ActivationFunction.RELU,
            output_activation_fn: ActivationFunction = ActivationFunction.NONE,
            p_dropout: Optional[float] = None):
        self.hidden_dims = hidden_dims
        self.hid_activation_fn = hid_activation_fn
        self.output_activation_fn = output_activation_fn
        self.p_dropout = p_dropout

    def create_predictor(self, input_dim: int, output_dim: int) -> TPredictor:
        return MultiLayerPerceptron(input_dim, output_dim, self.hidden_dims, hid_activation_fn=self.hid_activation_fn.get_torch_function(),
            output_activation_fn=self.output_activation_fn.get_torch_function(), p_dropout=self.p_dropout)


class RnnEncoderModule(torch.nn.Module):
    """
    Encodes a sequence of feature vectors, outputting a latent vector.
    The input sequence may either be fixed-length or variable-length.
    """

    class RnnType:
        GRU = "gru"
        """gated recurrent unit"""

        LSTM = "lstm"
        """long short-term memory"""

    def __init__(self, input_dim, latent_dim: int, rnn_type: RnnType = RnnType.LSTM):
        """
        :param input_dim: the input dimension per time slice
        :param latent_dim: the dimension of the latent output vector
        :param rnn_type: the type of recurrent network to use
        """
        super().__init__()
        self.window_dim_per_item = input_dim
        self.latent_dim = latent_dim
        self.rnn_type = rnn_type
        if rnn_type == self.RnnType.GRU:
            self.rnn = torch.nn.GRU(input_size=self.window_dim_per_item, hidden_size=latent_dim, batch_first=True)
        elif rnn_type == self.RnnType.LSTM:
            self.rnn = torch.nn.LSTM(input_size=self.window_dim_per_item, hidden_size=latent_dim, batch_first=True)
        else:
            raise ValueError(f"Unknown rnn type '{rnn_type}', use either 'gru' or 'lstm'")

    def __str__(self):
        return object_repr(self, dict(rnn=self.rnn))

    def forward(self, x: torch.Tensor, lengths: Optional[torch.Tensor] = None):
        """
        :param x: a tensor of size (batch_size, seq_length, dim_per_item)
        :param lengths: an optional tensor containing the lengths of the sequences; if None,
            all sequences are assumed to have the same full length
        :return: a tensor of size (batch_size, latent_dim)
        """
        if lengths is not None:
            x = torch.nn.utils.rnn.pack_padded_sequence(x, lengths, batch_first=True, enforce_sorted=False)
        if self.rnn_type == self.RnnType.GRU:
            _, l = self.rnn(x)
        elif self.rnn_type == self.RnnType.LSTM:
            _o, (l, _c) = self.rnn(x)
        else:
            raise ValueError(self.rnn_type)
        # l has shape (1, batch_size, latent_dim)
        l = l.squeeze(0)
        return l  # (batch_size, latent_dim)


class RnnEncoderFactory(EncoderFactory):
    def __init__(self,
            input_dim: int,
            latent_dim: int,
            rnn_type: RnnEncoderModule.RnnType = RnnEncoderModule.RnnType.GRU):
        self.input_dim = input_dim
        self.latent_dim = latent_dim
        self.rnn_type = rnn_type

    def create_encoder(self, input_dim: int, latent_dim: int):
        return RnnEncoderModule(input_dim, latent_dim, self.rnn_type)


class LSTNetworkEncoder(torch.nn.Module):
    """
    Adapts an LSTNetwork instance to the encoder interface
    """
    def __init__(self, lstnet: LSTNetwork):
        super().__init__()
        self.lstnet = lstnet

    def forward(self, x: torch.Tensor, lengths: Optional[torch.Tensor] = None):
        """
        :param x: a tensor of size (batch_size, seq_length, dim_per_item)
        :param lengths: an optional tensor containing the lengths of the sequences; if None,
            all sequences are assumed to have the same full length
        :return: a tensor of size (batch_size, latent_dim)
        """
        if lengths is not None:
            unique_lengths = torch.unique(lengths)
            if len(unique_lengths) != 1:
                raise ValueError("LSTNetwork does not support variable-length inputs")
        l = self.lstnet(x)
        return l


class LSTNetworkEncoderFactory(EncoderFactory):
    def __init__(self,
            num_input_time_slices: int,
            num_convolutions: int,
            num_cnn_time_slices: int,
            hid_rnn: int,
            skip: int,
            hid_skip: int,
            dropout: float = 0.2):
        self.dropout = dropout
        self.num_input_time_slices = num_input_time_slices
        self.num_convolutions = num_convolutions
        self.num_cnn_time_slices = num_cnn_time_slices
        self.hid_rnn = hid_rnn
        self.skip = skip
        self.hid_skip = hid_skip

    def create_encoder(self, input_dim: int, latent_dim: int) -> torch.nn.Module:
        lstnet = LSTNetwork(num_input_time_slices=self.num_input_time_slices,
            input_dim_per_time_slice=input_dim,
            num_convolutions=self.num_convolutions,
            num_cnn_time_slices=self.num_cnn_time_slices,
            hid_rnn=self.hid_rnn,
            skip=self.skip,
            hid_skip=self.hid_skip,
            dropout=self.dropout,
            mode=LSTNetwork.Mode.ENCODER)
        if lstnet.get_encoder_dim() != latent_dim:
            raise ValueError(f"LSTNetwork produces latent_dim={lstnet.get_encoder_dim()}; please adjust the parameter")
        return LSTNetworkEncoder(lstnet)

    def get_latent_dim(self) -> int:
        return LSTNetwork.compute_encoder_dim(self.hid_rnn, self.skip, self.hid_skip)


class SingleTargetDecoderModule(torch.nn.Module, DecoderProtocol):
    """
    Represents a decoder that output a single value for a single target item, taking as input the concatenation of the
    latent tensor (generated by the encoder) and the target item's feature vector.
    """
    def __init__(self, target_feature_dim, latent_dim, predictor_factory: PredictorFactory, output_dim=1):
        """
        :param target_feature_dim: the number of target item features
        :param latent_dim: the dimension of the latent vector generated by the encoder, which we receive as input
        :param predictor_factory: a factory for the creation of the predictor that will map the combined
            latent vector and target feature vector to the prediction of size `output_dim`
        :param output_dim: the output (prediction) dimension
        """
        super().__init__()
        self.target_feature_dim = target_feature_dim
        self.latent_dim = latent_dim
        self.predictor_input_dim = self.latent_dim + self.target_feature_dim
        self.predictor = predictor_factory.create_predictor(self.predictor_input_dim, output_dim)

    def __str__(self):
        return object_repr(self, dict(predictor=self.predictor))

    def forward(self, latent, target_features=None, target_lengths=None):
        if target_features is not None:
            # target_features must have shape (batch_size, 1, target_feature_dim)
            assert target_features.shape[1] == 1, "target_features must contain but one sequence item"
            target_features = torch.squeeze(target_features, 1)

            lf = torch.cat((latent, target_features), dim=1)
        else:
            lf = latent
        return self.predictor(lf)


class TargetSequenceDecoderModule(torch.nn.Module, DecoderProtocol, ToStringMixin):
    """
    Wrapper for decoders that take as input a latent representation (generated by an encoder)
    and a sequence of target features.
    It can generate either a single prediction for the entire sequence of target features or
    a sequence of predictions (one for each target sequence item), depending on the prediction/output mode.
    """
    class PredictionMode(Enum):
        """Defines how the prediction works"""

        SINGLE_LATENT = "single_latent"
        """
        Use an LSTM to process the target feature sequence and use only the final hidden state for prediction, 
        outputting a single average prediction only (for OutputMode.SINGLE_OUTPUT only)
        """

        MULTI_LATENT = "multi_latent"
        """Use an LSTM to process the target feature sequence and use all hidden states (full output) for prediction"""

        DIRECT = "direct"
        """Directly use the latent vector and target features to make predictions for each target sequence
        item (use with LatentPassOnMode.CONCAT_INPUT & NO_LATENT only)
        """

    class LatentPassOnMode(Enum):
        """Defines how the latent state from the encoder stage is passed on to the decoder"""

        INIT_HIDDEN = "init_hidden"
        """
        Pass on the encoder output as the initial hidden state of the LSTM 
        (only possible for OutputMode in {SINGLE_LATENT, MULTI_LATENT})
        """

        CONCAT_INPUT = "concat_input"
        """Pass on the encoder output by concatenating it with each target feature input vector"""

        NO_LATENT = "no_latent"
        """
        Do not pass on the latent vector at all (ignored by subsequent decoder component). 
        This is mostly useful for ablation testing.
        """

    class OutputMode(Enum):
        """Defines how to treat multiple predictions (for PredictionMode != SINGLE_LATENT)"""

        SINGLE_OUTPUT = "single"
        """Output a single result from a single input (for PredictionMode.SINGLE_LATENT only)"""

        SINGLE_OUTPUT_MEAN = "mean"
        """Output the mean of multiple (intermediate) predictions"""

        MULTI_OUTPUT = "multi"
        """Output multiple predictions directly"""

    def __init__(self,
            target_feature_dim: int,
            latent_dim: int,
            predictor_factory: PredictorFactory,
            output_dim: int = 1,
            prediction_mode: PredictionMode = PredictionMode.MULTI_LATENT,
            latent_pass_on_mode: LatentPassOnMode = LatentPassOnMode.CONCAT_INPUT,
            output_mode: OutputMode = OutputMode.MULTI_OUTPUT,
            p_recurrent_dropout: float = 0.0):
        super().__init__()
        if not ((prediction_mode == self.PredictionMode.SINGLE_LATENT) ==
                (output_mode == self.OutputMode.SINGLE_OUTPUT)):  # SINGLE_LATENT <=> SINGLE_OUTPUT
            raise ValueError(f"{self.PredictionMode.SINGLE_LATENT} must coincide with {self.OutputMode.SINGLE_OUTPUT}; "
                             f"got {prediction_mode} and {output_mode}")
        if prediction_mode == self.PredictionMode.DIRECT and \
                latent_pass_on_mode not in (self.LatentPassOnMode.CONCAT_INPUT, self.LatentPassOnMode.NO_LATENT):
            raise ValueError(f"{prediction_mode} requires {self.LatentPassOnMode.CONCAT_INPUT}")
        if latent_pass_on_mode == self.LatentPassOnMode.INIT_HIDDEN and \
                prediction_mode not in (self.PredictionMode.SINGLE_LATENT, self.PredictionMode.MULTI_LATENT):
            raise ValueError(f"{output_mode} requires {self.PredictionMode.SINGLE_LATENT} or {self.PredictionMode.MULTI_LATENT} ")
        if latent_pass_on_mode == self.LatentPassOnMode.NO_LATENT:
            latent_dim = 0
        self.latent_pass_on_mode = latent_pass_on_mode
        self.prediction_mode = prediction_mode
        self.output_mode = output_mode
        self.target_feature_dim = target_feature_dim
        self.latent_dim = latent_dim
        if prediction_mode == self.PredictionMode.DIRECT:
            self.lstm = None
            predictor_input_dim = self.latent_dim + self.target_feature_dim
        else:
            if latent_pass_on_mode == self.LatentPassOnMode.INIT_HIDDEN:
                rnn_input_dim = target_feature_dim
            else:
                rnn_input_dim = target_feature_dim + latent_dim
            self.lstm = torch.nn.LSTM(rnn_input_dim, self.latent_dim, batch_first=True, dropout=p_recurrent_dropout)
            predictor_input_dim = self.latent_dim
        self.predictor = predictor_factory.create_predictor(predictor_input_dim, output_dim)

    def _tostring_exclude_private(self) -> bool:
        return True

    def forward(self, latent, target_features=None, target_lengths=None):
        """
        :param latent: a tensor of shape (batch_size, latent_dim)
        :param target_features: a tensor of shape (batch_size, max_seq_length, target_feature_dim)
        :param target_lengths: a tensor indicating the lengths of the sequences in target_features
        :return:
        """
        if target_features is None:
            raise ValueError(f"target_features cannot be None when using {self.__class__}")

        # latent has shape (batch_size, latentDim)
        # targetFeatures has shape (batch_size, maxSeqLength, targetFeatureDim)
        batch_size = target_features.shape[0]
        use_lstm = self.prediction_mode != self.PredictionMode.DIRECT

        lstm_input, s0, latent_plus_target_features = None, None, None
        if self.latent_pass_on_mode == self.LatentPassOnMode.INIT_HIDDEN:
            if target_lengths is not None:
                lstm_input = torch.nn.utils.rnn.pack_padded_sequence(target_features, target_lengths, batch_first=True,
                    enforce_sorted=False)
            else:
                lstm_input = target_features
            c0 = torch.zeros((1, batch_size, self.latent_dim)).to(latent.device)
            h0 = latent.unsqueeze(0)
            s0 = (h0, c0)
        elif self.latent_pass_on_mode in (self.LatentPassOnMode.CONCAT_INPUT, self.LatentPassOnMode.NO_LATENT):
            if self.latent_pass_on_mode == self.LatentPassOnMode.NO_LATENT:
                latent_plus_target_features = target_features
            else:
                latent = latent.unsqueeze(1)  # (batch_size, 1, latentDim)
                latent = latent.expand(-1, target_features.shape[1], -1)  # (batch_size, maxSeqLength, latentDim)
                latent_plus_target_features = torch.cat((latent, target_features), dim=2)
                # latent_plus_target_features has shape (batch_size, maxSeqLength, latentDim + targetFeatureDim)
            if use_lstm:
                lstm_input = torch.nn.utils.rnn.pack_padded_sequence(latent_plus_target_features, target_lengths, batch_first=True,
                    enforce_sorted=False)
                s0 = None
        else:
            raise ValueError(f"Unknown latent pass-on mode '{self.latent_pass_on_mode}'")

        if self.prediction_mode == self.PredictionMode.SINGLE_LATENT:  # use only final latent state and produce a single output
            _, (hn, _) = self.lstm(lstm_input, s0)
            encoding = hn.squeeze(0)
            result = self.predictor(encoding)
        else:  # compute multiple predictions (and optionally compute their mean)
            if self.prediction_mode == self.PredictionMode.MULTI_LATENT:  # use all latent states
                hseq, _ = self.lstm(lstm_input, s0)
                encodings, lengths = torch.nn.utils.rnn.pad_packed_sequence(hseq, batch_first=True)  # (batch_size, maxSeqLength, latentDim)
                predictions = self.predictor(encodings)  # (batch_size, maxSeqLength, outputDim)
            elif self.prediction_mode == self.PredictionMode.DIRECT:  # directly map concatenated values to outputs via predictor
                predictions = self.predictor(latent_plus_target_features)
            else:
                raise ValueError(f"Unknown prediction mode '{self.prediction_mode}'")
            if self.output_mode == self.OutputMode.SINGLE_OUTPUT_MEAN:
                mean_predictions = predictions.data.new(batch_size, 1)
                for i, l in enumerate(target_lengths):
                    mean_predictions[i] = predictions[i][:l].sum() / l
                result = mean_predictions
            elif self.output_mode == self.OutputMode.MULTI_OUTPUT:
                return predictions
            else:
                raise ValueError(self.output_mode)

        return result


class TargetSequenceDecoderFactory(DecoderFactory):
    """
    A factory for :class:`TargetSequenceDecoderModule` which takes the latent encoding and a sequence of target
    items as input
    """
    def __init__(self,
            prediction_mode: TargetSequenceDecoderModule.PredictionMode = TargetSequenceDecoderModule.PredictionMode.MULTI_LATENT,
            output_mode: TargetSequenceDecoderModule.OutputMode = TargetSequenceDecoderModule.OutputMode.MULTI_OUTPUT,
            latent_pass_on_mode: TargetSequenceDecoderModule.LatentPassOnMode = TargetSequenceDecoderModule.LatentPassOnMode.CONCAT_INPUT,
            predictor_factory: Optional[PredictorFactory] = None,
            p_recurrent_dropout: float = 0.0,
            output_dim: int = 1):
        if predictor_factory is None:
            predictor_factory = LinearPredictorFactory()
        self.output_dim = output_dim
        self.p_recurrent_dropout = p_recurrent_dropout
        self.prediction_mode = prediction_mode
        self.output_mode = output_mode
        self.latent_pass_on_mode = latent_pass_on_mode
        self.predictor_factory = predictor_factory

    def create_decoder(self, latent_dim: int, target_feature_dim: int) -> torch.nn.Module:
        return TargetSequenceDecoderModule(target_feature_dim, latent_dim, self.predictor_factory,
            prediction_mode=self.prediction_mode,
            output_mode=self.output_mode,
            latent_pass_on_mode=self.latent_pass_on_mode,
            output_dim=self.output_dim,
            p_recurrent_dropout=self.p_recurrent_dropout)


class SingleTargetDecoderFactory(DecoderFactory):
    """
    A factory for :class:`SingleTargetDecoderModule` which takes the latent encoding and a single-element sequence of target
    items as input, producing a single prediction
    """
    def __init__(self, predictor_factory: PredictorFactory):
        self.predictor_factory = predictor_factory

    def create_decoder(self, latent_dim: int, target_feature_dim: int) -> torch.nn.Module:
        return SingleTargetDecoderModule(target_feature_dim, latent_dim, self.predictor_factory)


class EncoderDecoderModule(torch.nn.Module):
    """
    Represents and encoder-decoder (where both components can be injected). It takes a history sequence and a sequence of
    target feature vectors as input. Both sequences are potentially of variable length, and for the target sequence,
    the common special case where there is but one target and thus one prediction to be made is specifically catered for
    using dedicated decoders (see :class:`SingleTargetDecoderModule`).

    The module first encodes the history sequence to a latent vector and then uses the decoder to map this latent vector
    along with the target features to a prediction.
    """
    def __init__(self, encoder: TEncoder, decoder: TDecoder, variable_history_length: bool):
        """
        :param encoder: a torch module satisfying :class:`EncoderProtocol`
        :param decoder: a torch module satisfying :class:`DecoderProtocol`
        :param variable_history_length: whether the history sequence is variable-length.
            If it is not, then the model will not pass on the lengths tensor to the encoder, allowing it to simplify
            its handling of this case (even if the original input provides the lengths).
        """
        super().__init__()
        self.variable_history_length = variable_history_length
        self.encoder = encoder
        self.decoder = decoder

    def __str__(self):
        return object_repr(self, dict(encoder=self.encoder, predictor=self.decoder))

    def forward(self, window_features: torch.Tensor,
            window_lengths: Optional[torch.Tensor] = None,
            target_features: Optional[torch.Tensor] = None,
            target_lengths: Optional[torch.Tensor] = None):
        """
        :param window_features: a tensor of size (batch_size, max(window_lengths), dim_per_window_item) containing the window features
        :param window_lengths: a tensor containing the lengths of windows in `w`
        :param target_features: an optional tensor containing target features with shape
            (batch_size, max_target_seq_length, target_feature_dim).
            For the case where there is only one target item (no actual sequence), `max_target_seq_length`
            should be 1.
        :param target_lengths: an optional tensor containing the lengths target the target sequences, allowing
            the actual sequence lengths to differ
        """
        if self.variable_history_length:
            latent = self.encoder(window_features, window_lengths)
        else:
            latent = self.encoder(window_features)
        return self.decoder(latent, target_features, target_lengths)



import atexit
import enum
import glob
import logging
import os
import pickle
import re
import sqlite3
import threading
import time
from abc import abstractmethod, ABC
from collections import OrderedDict
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Iterator, List, Optional, TypeVar, Generic, Union, Hashable

from .hash import pickle_hash
from .pickle import load_pickle, dump_pickle, setstate

log = logging.getLogger(__name__)

T = TypeVar("T")
TKey = TypeVar("TKey")
THashableKey = TypeVar("THashableKey", bound=Hashable)
TValue = TypeVar("TValue")
TData = TypeVar("TData")


class BoxedValue(Generic[TValue]):
    """
    Container for a value, which can be used in caches where values may be None (to differentiate the value not being present in the cache
    from the cached value being None)
    """
    def __init__(self, value: TValue):
        self.value = value


class KeyValueCache(Generic[TKey, TValue], ABC):
    @abstractmethod
    def set(self, key: TKey, value: TValue):
        """
        Sets a cached value

        :param key: the key under which to store the value
        :param value: the value to store; since None is used indicate the absence of a value, None should not be
            used a value
        """
        pass

    @abstractmethod
    def get(self, key: TKey) -> Optional[TValue]:
        """
        Retrieves a cached value

        :param key: the lookup key
        :return: the cached value or None if no value is found
        """
        pass


class InMemoryKeyValueCache(KeyValueCache[TKey, TValue], Generic[TKey, TValue]):
    """A simple in-memory cache (which uses a dictionary internally).

    This class can be instantiated directly, but for better typing support, one can instead
    inherit from it and provide the types of the key and value as type arguments. For example for
    a cache with string keys and integer values:

    .. code-block:: python

        class MyCache(InMemoryKeyValueCache[str, int]):
            pass
    """
    def __init__(self):
        self.cache = {}

    def set(self, key: TKey, value: TValue):
        self.cache[key] = value

    def get(self, key: TKey) -> Optional[TValue]:
        return self.cache.get(key)

    def empty(self):
        self.cache = {}

    def __len__(self):
        return len(self.cache)



# mainly kept as a marker and for backwards compatibility, but may be extended in the future
class PersistentKeyValueCache(KeyValueCache[TKey, TValue], Generic[TKey, TValue], ABC):
    pass


class PersistentList(Generic[TValue], ABC):
    @abstractmethod
    def append(self, item: TValue):
        """
        Adds an item to the cache

        :param item: the item to store
        """
        pass

    @abstractmethod
    def iter_items(self) -> Iterator[TValue]:
        """
        Iterates over the items in the persisted list

        :return: generator of item
        """
        pass


class DelayedUpdateHook:
    """
    Ensures that a given function is executed after an update happens, but delay the execution until
    there are no further updates for a certain time period
    """
    def __init__(self, fn: Callable[[], Any], time_period_secs, periodically_executed_fn: Optional[Callable[[], Any]] = None):
        """
        :param fn: the function to eventually call after an update
        :param time_period_secs: the time that must pass while not receiving further updates for fn to be called
        :param periodically_executed_fn: a function to execute periodically (every timePeriodSecs seconds) in the busy waiting loop,
            which may, for example, log information or apply additional executions, which must not interfere with the correctness of
            the execution of fn
        """
        self.periodicallyExecutedFn = periodically_executed_fn
        self.fn = fn
        self.timePeriodSecs = time_period_secs
        self._lastUpdateTime = None
        self._thread = None
        self._threadLock = threading.Lock()

    def handle_update(self):
        """
        Notifies of an update and ensures that the function passed at construction is eventually called
        (after no more updates are received within the respective time window)
        """
        self._lastUpdateTime = time.time()

        def do_periodic_check():
            while True:
                time.sleep(self.timePeriodSecs)
                time_passed_since_last_update = time.time() - self._lastUpdateTime
                if self.periodicallyExecutedFn is not None:
                    self.periodicallyExecutedFn()
                if time_passed_since_last_update >= self.timePeriodSecs:
                    self.fn()
                    return

        # noinspection DuplicatedCode
        if self._thread is None or not self._thread.is_alive():
            self._threadLock.acquire()
            if self._thread is None or not self._thread.is_alive():
                self._thread = threading.Thread(target=do_periodic_check, daemon=False)
                self._thread.start()
            self._threadLock.release()


class PeriodicUpdateHook:
    """
    Periodically checks whether a function shall be called as a result of an update, the function potentially
    being non-atomic (i.e. it may take a long time to execute such that new updates may come in while it is
    executing). Two function all mechanisms are in place:

        * a function which is called if there has not been a new update for a certain time period (which may be called
          several times if updates come in while the function is being executed)
        * a function which is called periodically

    """
    def __init__(self, check_interval_secs: float, no_update_time_period_secs: float = None, no_update_fn: Callable[[], Any] = None,
            periodic_fn: Optional[Callable[[], Any]] = None):
        """
        :param check_interval_secs: the time period, in seconds, between checks
        :param no_update_time_period_secs: the time period after which to execute noUpdateFn if no further updates have come in.
            This must be at least as large as checkIntervalSecs. If None, use checkIntervalSecs.
        :param no_update_fn: the function to call if there have been no further updates for noUpdateTimePeriodSecs seconds
        :param periodic_fn: a function to execute periodically (every checkIntervalSecs seconds) in the busy waiting loop,
            which may, for example, log information or apply additional executions, which must not interfere with the correctness of
            the execution of fn
        """
        if no_update_time_period_secs is None:
            no_update_time_period_secs = check_interval_secs
        elif no_update_time_period_secs < check_interval_secs:
            raise ValueError("noUpdateTimePeriodSecs must be at least as large as checkIntervalSecs")
        self._periodic_fn = periodic_fn
        self._check_interval_secs = check_interval_secs
        self._no_update_time_period_secs = no_update_time_period_secs
        self._no_update_fn = no_update_fn
        self._last_update_time = None
        self._thread = None
        self._thread_lock = threading.Lock()

    def handle_update(self):
        """
        Notifies of an update, making sure the functions given at construction will be called as specified
        """
        self._last_update_time = time.time()

        def do_periodic_check():
            while True:
                time.sleep(self._check_interval_secs)
                check_time = time.time()
                if self._periodic_fn is not None:
                    self._periodic_fn()
                time_passed_since_last_update = check_time - self._last_update_time
                if time_passed_since_last_update >= self._no_update_time_period_secs:
                    if self._no_update_fn is not None:
                        self._no_update_fn()
                    # if no further updates have come in, we terminate the thread
                    if self._last_update_time < check_time:
                        return

        # noinspection DuplicatedCode
        if self._thread is None or not self._thread.is_alive():
            self._thread_lock.acquire()
            if self._thread is None or not self._thread.is_alive():
                self._thread = threading.Thread(target=do_periodic_check, daemon=False)
                self._thread.start()
            self._thread_lock.release()


class PicklePersistentKeyValueCache(PersistentKeyValueCache[TKey, TValue]):
    """
    Represents a key-value cache as a dictionary which is persisted in a file using pickle
    """
    def __init__(self, pickle_path, version=1, save_on_update=True, deferred_save_delay_secs=1.0):
        """
        :param pickle_path: the path of the file where the cache values are to be persisted
        :param version: the version of cache entries. If a persisted cache with a non-matching version is found,
            it is discarded
        :param save_on_update: whether to persist the cache after an update; the cache is saved in a deferred
            manner and will be saved after deferredSaveDelaySecs if no new updates have arrived in the meantime,
            i.e. it will ultimately be saved deferredSaveDelaySecs after the latest update
        :param deferred_save_delay_secs: the number of seconds to wait for additional data to be added to the cache
            before actually storing the cache after a cache update
        """
        self.deferred_save_delay_secs = deferred_save_delay_secs
        self.pickle_path = pickle_path
        self.version = version
        self.save_on_update = save_on_update
        cache_found = False
        if os.path.exists(pickle_path):
            try:
                log.info(f"Loading cache from {pickle_path}")
                persisted_version, self.cache = load_pickle(pickle_path)
                if persisted_version == version:
                    cache_found = True
            except EOFError:
                log.warning(f"The cache file in {pickle_path} is corrupt")
        if not cache_found:
            self.cache = {}
        self._update_hook = DelayedUpdateHook(self.save, deferred_save_delay_secs)
        self._write_lock = threading.Lock()

    def save(self):
        """
        Saves the cache in the file whose path was provided at construction
        """
        with self._write_lock:  # avoid concurrent modification while saving
            log.info(f"Saving cache to {self.pickle_path}")
            dump_pickle((self.version, self.cache), self.pickle_path)

    def get(self, key: TKey) -> Optional[TValue]:
        return self.cache.get(key)

    def set(self, key: TKey, value: TValue):
        with self._write_lock:
            self.cache[key] = value
            if self.save_on_update:
                self._update_hook.handle_update()


class SlicedPicklePersistentList(PersistentList):
    """
    Object handling the creation and access to sliced pickle caches
    """
    def __init__(self, directory, pickle_base_name, num_entries_per_slice=100000):
        """
        :param directory: path to the directory where the sliced caches are to be stored
        :param pickle_base_name: base name for the pickle, where slices will have the names {pickleBaseName}_sliceX.pickle
        :param num_entries_per_slice: how many entries should be stored in each cache
        """
        self.directory = directory
        self.pickleBaseName = pickle_base_name
        self.numEntriesPerSlice = num_entries_per_slice

        # Set up the variables for the sliced cache
        self.slice_id = 0
        self.index_in_slice = 0
        self.cache_of_slice = []

        # Search directory for already present sliced caches
        self.slicedFiles = self._find_sliced_caches()

        # Helper variable to ensure object is only modified within a with-clause
        self._currentlyInWithClause = False

    def __enter__(self):
        self._currentlyInWithClause = True
        if self.cache_exists():
            # Reset state to enable the appending of more items to the cache
            self._set_last_cache_state()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._dump()
        self._currentlyInWithClause = False

    def append(self, item):
        """
        Append item to cache
        :param item: entry in the cache
        """
        if not self._currentlyInWithClause:
            raise Exception("Class needs to be instantiated within a with-clause to ensure correct storage")

        if (self.index_in_slice + 1) % self.numEntriesPerSlice == 0:
            self._dump()

        self.cache_of_slice.append(item)
        self.index_in_slice += 1

    def iter_items(self) -> Iterator[Any]:
        """
        Iterate over entries in the sliced cache
        :return: iterator over all items in the cache
        """
        for filePath in self.slicedFiles:
            log.info(f"Loading sliced pickle list from {filePath}")
            cached_pickle = self._load_pickle(filePath)
            for item in cached_pickle:
                yield item

    def clear(self):
        """
        Clears the cache if it exists
        """
        if self.cache_exists():
            for filePath in self.slicedFiles:
                os.unlink(filePath)

    def cache_exists(self) -> bool:
        """
        Does this cache already exist
        :return: True if cache exists, False if not
        """
        return len(self.slicedFiles) > 0

    def _set_last_cache_state(self):
        """
        Sets the state so as to be able to add items to an existing cache
        """
        log.info("Resetting last state of cache...")
        self.slice_id = len(self.slicedFiles) - 1
        self.cache_of_slice = self._load_pickle(self._pickle_path(self.slice_id))
        self.index_in_slice = len(self.cache_of_slice) - 1
        if self.index_in_slice >= self.numEntriesPerSlice:
            self._next_slice()

    def _dump(self):
        """
        Dumps the current cache (if non-empty)
        """
        if len(self.cache_of_slice) > 0:
            pickle_path = self._pickle_path(str(self.slice_id))
            log.info(f"Saving sliced cache to {pickle_path}")
            dump_pickle(self.cache_of_slice, pickle_path)
            self.slicedFiles.append(pickle_path)

            # Update slice number and reset indexing and cache
            self._next_slice()
        else:
            log.warning("Unexpected behavior: Dump was called when cache of slice is 0!")

    def _next_slice(self):
        """
        Updates sliced cache state for the next slice
        """
        self.slice_id += 1
        self.index_in_slice = 0
        self.cache_of_slice = []

    def _find_sliced_caches(self) -> List[str]:
        """
        Finds all pickled slices associated with this cache
        :return: list of sliced pickled files
        """
        # glob.glob permits the usage of unix-style pathnames matching. (below we find all ..._slice*.pickle files)
        list_of_file_names = glob.glob(self._pickle_path("*"))
        # Sort the slices to ensure it is in the same order as they was produced (regex replaces everything not a number with empty string).
        list_of_file_names.sort(key=lambda f: int(re.sub(r'\D', '', f)))
        return list_of_file_names

    def _load_pickle(self, pickle_path: str) -> List[Any]:
        """
        Loads pickle if file path exists, and persisted version is correct.
        :param pickle_path: file path
        :return: list with objects
        """
        cached_pickle = []
        if os.path.exists(pickle_path):
            try:
                cached_pickle = load_pickle(pickle_path)
            except EOFError:
                log.warning(f"The cache file in {pickle_path} is corrupt")
        else:
            raise Exception(f"The file {pickle_path} does not exist!")
        return cached_pickle

    def _pickle_path(self, slice_suffix) -> str:
        return f"{os.path.join(self.directory, self.pickleBaseName)}_slice{slice_suffix}.pickle"


class SqliteConnectionManager:
    _connections: List[sqlite3.Connection] = []
    _atexit_handler_registered = False

    @classmethod
    def _register_at_exit_handler(cls):
        if not cls._atexit_handler_registered:
            cls._atexit_handler_registered = True
            atexit.register(cls._cleanup)

    @classmethod
    def open_connection(cls, path):
        cls._register_at_exit_handler()
        conn = sqlite3.connect(path, check_same_thread=False)
        cls._connections.append(conn)
        return conn

    @classmethod
    def _cleanup(cls):
        for conn in cls._connections:
            conn.close()
        cls._connections = []


class SqlitePersistentKeyValueCache(PersistentKeyValueCache[TKey, TValue]):
    class KeyType(enum.Enum):
        STRING = ("VARCHAR(%d)", )
        INTEGER = ("LONG", )

    def __init__(self, path, table_name="cache", deferred_commit_delay_secs=1.0, key_type: KeyType = KeyType.STRING,
            max_key_length=255):
        """
        :param path: the path to the file that is to hold the SQLite database
        :param table_name: the name of the table to create in the database
        :param deferred_commit_delay_secs: the time frame during which no new data must be added for a pending transaction to be committed
        :param key_type: the type to use for keys; for complex keys (i.e. tuples), use STRING (conversions to string are automatic)
        :param max_key_length: the maximum key length for the case where the key_type can be parametrised (e.g. STRING)
        """
        self.path = path
        self.conn = SqliteConnectionManager.open_connection(path)
        self.table_name = table_name
        self.max_key_length = 255
        self.key_type = key_type
        self._update_hook = DelayedUpdateHook(self._commit, deferred_commit_delay_secs)
        self._num_entries_to_be_committed = 0
        self._conn_mutex = threading.Lock()

        cursor = self.conn.cursor()
        cursor.execute(f"SELECT name FROM sqlite_master WHERE type='table';")
        if table_name not in [r[0] for r in cursor.fetchall()]:
            log.info(f"Creating cache table '{self.table_name}' in {path}")
            key_db_type = key_type.value[0]
            if "%d" in key_db_type:
                key_db_type = key_db_type % max_key_length
            cursor.execute(f"CREATE TABLE {table_name} (cache_key {key_db_type} PRIMARY KEY, cache_value BLOB);")
        cursor.close()

    def _key_db_value(self, key):
        if self.key_type == self.KeyType.STRING:
            s = str(key)
            if len(s) > self.max_key_length:
                raise ValueError(f"Key too long, maximal key length is {self.max_key_length}")
            return s
        elif self.key_type == self.KeyType.INTEGER:
            return int(key)
        else:
            raise Exception(f"Unhandled key type {self.key_type}")

    def _commit(self):
        self._conn_mutex.acquire()
        try:
            log.info(f"Committing {self._num_entries_to_be_committed} cache entries to the SQLite database {self.path}")
            self.conn.commit()
            self._num_entries_to_be_committed = 0
        finally:
            self._conn_mutex.release()

    def set(self, key: TKey, value: TValue):
        self._conn_mutex.acquire()
        try:
            cursor = self.conn.cursor()
            key = self._key_db_value(key)
            cursor.execute(f"SELECT COUNT(*) FROM {self.table_name} WHERE cache_key=?", (key,))
            if cursor.fetchone()[0] == 0:
                cursor.execute(f"INSERT INTO {self.table_name} (cache_key, cache_value) VALUES (?, ?)",
                               (key, pickle.dumps(value)))
            else:
                cursor.execute(f"UPDATE {self.table_name} SET cache_value=? WHERE cache_key=?", (pickle.dumps(value), key))
            self._num_entries_to_be_committed += 1
            cursor.close()
        finally:
            self._conn_mutex.release()

        self._update_hook.handle_update()

    def _execute(self, cursor, *query):
        try:
            cursor.execute(*query)
        except sqlite3.DatabaseError as e:
            raise Exception(f"Error executing query for {self.path}: {e}")

    def get(self, key: TKey) -> Optional[TValue]:
        self._conn_mutex.acquire()
        try:
            cursor = self.conn.cursor()
            key = self._key_db_value(key)
            self._execute(cursor, f"SELECT cache_value FROM {self.table_name} WHERE cache_key=?", (key,))
            row = cursor.fetchone()
            cursor.close()
            if row is None:
                return None
            return pickle.loads(row[0])
        finally:
            self._conn_mutex.release()

    def __len__(self):
        self._conn_mutex.acquire()
        try:
            cursor = self.conn.cursor()
            cursor.execute(f"SELECT COUNT(*) FROM {self.table_name}")
            cnt = cursor.fetchone()[0]
            cursor.close()
            return cnt
        finally:
            self._conn_mutex.release()

    def iter_items(self):
        self._conn_mutex.acquire()
        try:
            cursor = self.conn.cursor()
            cursor.execute(f"SELECT cache_key, cache_value FROM {self.table_name}")
            while True:
                row = cursor.fetchone()
                if row is None:
                    break
                yield row[0], pickle.loads(row[1])
            cursor.close()
        finally:
            self._conn_mutex.release()


class SqlitePersistentList(PersistentList):
    def __init__(self, path):
        self.keyValueCache = SqlitePersistentKeyValueCache(path, key_type=SqlitePersistentKeyValueCache.KeyType.INTEGER)
        self.nextKey = len(self.keyValueCache)

    def append(self, item):
        self.keyValueCache.set(self.nextKey, item)
        self.nextKey += 1

    def iter_items(self):
        for item in self.keyValueCache.iter_items():
            yield item[1]


class CachedValueProviderMixin(Generic[TKey, TValue, TData], ABC):
    """
    Represents a value provider that can provide values associated with (hashable) keys via a cache or, if
    cached values are not yet present, by computing them.
    """
    def __init__(self, cache: Optional[KeyValueCache[TKey, TValue]] = None,
            cache_factory: Optional[Callable[[], KeyValueCache[TKey, TValue]]] = None, persist_cache=False, box_values=False):
        """
        :param cache: the cache to use or None. If None, caching will be disabled
        :param cache_factory: a factory with which to create the cache (or recreate it after unpickling if `persistCache` is False, in which
            case this factory must be picklable)
        :param persist_cache: whether to persist the cache when pickling
        :param box_values: whether to box values, such that None is admissible as a value
        """
        self._persistCache = persist_cache
        self._boxValues = box_values
        self._cache = cache
        self._cacheFactory = cache_factory
        if self._cache is None and cache_factory is not None:
            self._cache = cache_factory()

    def __getstate__(self):
        if not self._persistCache:
            d = self.__dict__.copy()
            d["_cache"] = None
            return d
        return self.__dict__

    def __setstate__(self, state):
        setstate(CachedValueProviderMixin, self, state, renamed_properties={"persistCache": "_persistCache"})
        if not self._persistCache and self._cacheFactory is not None:
            self._cache = self._cacheFactory()

    def _provide_value(self, key, data: Optional[TData] = None):
        """
        Provides the value for the key by retrieving the associated value from the cache or, if no entry in the
        cache is found, by computing the value via _computeValue

        :param key: the key for which to provide the value
        :param data: optional data required to compute a value
        :return: the retrieved or computed value
        """
        if self._cache is None:
            return self._compute_value(key, data)
        value = self._cache.get(key)
        if value is None:
            value = self._compute_value(key, data)
            self._cache.set(key, value if not self._boxValues else BoxedValue(value))
        else:
            if self._boxValues:
                value: BoxedValue[TValue]
                value = value.value
        return value

    @abstractmethod
    def _compute_value(self, key: TKey, data: Optional[TData]) -> TValue:
        """
        Computes the value for the given key

        :param key: the key for which to compute the value
        :return: the computed value
        """
        pass


def cached(fn: Callable[[], T], pickle_path, function_name=None, validity_check_fn: Optional[Callable[[T], bool]] = None,
        backend="pickle", protocol=pickle.HIGHEST_PROTOCOL, load=True, version=None) -> T:
    """
    Calls the given function unless its result is already cached (in a pickle), in which case it will read the cached result
    and return it.

    Rather than directly calling this function, consider using the decorator variant :func:`pickle_cached`.

    :param fn: the function whose result is to be cached
    :param pickle_path: the path in which to store the cached result
    :param function_name: the name of the function fn (for the case where its __name__ attribute is not
        informative)
    :param validity_check_fn: an optional function to call in order to check whether a cached result is still valid;
        the function shall return True if the result is still valid and false otherwise. If a cached result is invalid,
        the function fn is called to compute the result and the cached result is updated.
    :param backend: pickle or joblib
    :param protocol: the pickle protocol version
    :param load: whether to load a previously persisted result; if False, do not load an old result but store the newly computed result
    :param version: if not None, previously persisted data will only be returned if it was stored with the same version
    :return: the result (either obtained from the cache or the function)
    """
    if function_name is None:
        function_name = fn.__name__

    def call_fn_and_cache_result():
        res = fn()
        log.info(f"Saving cached result in {pickle_path}")
        if version is not None:
            persisted_res = {"__cacheVersion": version, "obj": res}
        else:
            persisted_res = res
        dump_pickle(persisted_res, pickle_path, backend=backend, protocol=protocol)
        return res

    if os.path.exists(pickle_path):
        if load:
            log.info(f"Loading cached result of function '{function_name}' from {pickle_path}")
            result = load_pickle(pickle_path, backend=backend)
            if validity_check_fn is not None:
                if not validity_check_fn(result):
                    log.info(f"Cached result is no longer valid, recomputing ...")
                    result = call_fn_and_cache_result()
            if version is not None:
                cached_version = None
                if type(result) == dict:
                    cached_version = result.get("__cacheVersion")
                if cached_version != version:
                    log.info(f"Cached result has incorrect version ({cached_version}, expected {version}), recomputing ...")
                    result = call_fn_and_cache_result()
                else:
                    result = result["obj"]
            return result
        else:
            log.info(f"Ignoring previously stored result in {pickle_path}, calling function '{function_name}' ...")
            return call_fn_and_cache_result()
    else:
        log.info(f"No cached result found in {pickle_path}, calling function '{function_name}' ...")
        return call_fn_and_cache_result()


def pickle_cached(cache_base_path: str, filename_prefix: str = None, filename: str = None, backend="pickle",
        protocol=pickle.HIGHEST_PROTOCOL, load=True, version=None):
    """
    Function decorator for caching function results via pickle.

    Add this decorator to any function to cache its results in pickle files.
    The function may have arguments, in which case the cache will be specific to the actual arguments
    by computing a hash code from their pickled representation.

    :param cache_base_path: the directory where the pickle cache file will be stored
    :param filename_prefix: a prefix of the name of the cache file to be created, to which the function name and, where applicable,
        a hash code of the function arguments as well as the extension ".cache.pickle" will be appended.
        The prefix need not end in a separator, as "-" will automatically be added between filename components.
    :param filename: the full file name of the cache file to be created; if the function takes arguments, the filename must
        contain a placeholder '%s' for the argument hash
    :param backend: the serialisation backend to use (see dumpPickle)
    :param protocol: the pickle protocol version to use
    :param load: whether to load a previously persisted result; if False, do not load an old result but store the newly computed result
    :param version: if not None, previously persisted data will only be returned if it was stored with the same version
    """
    os.makedirs(cache_base_path, exist_ok=True)

    if filename_prefix is None:
        filename_prefix = ""
    else:
        filename_prefix += "-"

    def decorator(fn: Callable, *_args, **_kwargs):

        @wraps(fn)
        def wrapped(*args, **kwargs):
            hash_code_str = None
            have_args = len(args) > 0 or len(kwargs) > 0
            if have_args:
                hash_code_str = pickle_hash((args, kwargs))
            if filename is None:
                pickle_filename = filename_prefix + fn.__qualname__.replace(".<locals>.", ".")
                if hash_code_str is not None:
                    pickle_filename += "-" + hash_code_str
                pickle_filename += ".cache.pickle"
            else:
                if hash_code_str is not None:
                    if "%s" not in filename:
                        raise Exception("Function called with arguments but full cache filename contains no placeholder (%s) "
                                        "for argument hash")
                    pickle_filename = filename % hash_code_str
                else:
                    if "%s" in filename:
                        raise Exception("Function without arguments but full cache filename with placeholder (%s) was specified")
                    pickle_filename = filename
            pickle_path = os.path.join(cache_base_path, pickle_filename)
            return cached(lambda: fn(*args, **kwargs), pickle_path, function_name=fn.__name__, backend=backend, load=load,
                version=version, protocol=protocol)

        return wrapped

    return decorator


PickleCached = pickle_cached  # for backward compatibility


class LoadSaveInterface(ABC):
    @abstractmethod
    def save(self, path: str) -> None:
        pass

    @classmethod
    @abstractmethod
    def load(cls: T, path: str) -> T:
        pass


class PickleLoadSaveMixin(LoadSaveInterface):
    def save(self, path: Union[str, Path], backend="pickle"):
        """
        Saves the instance as pickle

        :param path:
        :param backend: pickle, cloudpickle, or joblib
        """
        dump_pickle(self, path, backend=backend)

    @classmethod
    def load(cls, path: Union[str, Path], backend="pickle"):
        """
        Loads a class instance from pickle

        :param path:
        :param backend: pickle, cloudpickle, or joblib
        :return: instance of the present class
        """
        log.info(f"Loading instance of {cls} from {path}")
        result = load_pickle(path, backend=backend)
        if not isinstance(result, cls):
            raise Exception(f"Excepted instance of {cls}, instead got: {result.__class__.__name__}")
        return result


class LRUCache(KeyValueCache[THashableKey, TValue], Generic[THashableKey, TValue]):
    def __init__(self, capacity: int) -> None:
        self._cache = OrderedDict()
        self._capacity = capacity

    def get(self, key: THashableKey) -> TValue:
        if key not in self._cache:
            return None
        self._cache.move_to_end(key)
        return self._cache[key]

    def set(self, key: THashableKey, value: TValue):
        if key in self._cache:
            self._cache.move_to_end(key)
        self._cache[key] = value
        if len(self._cache) > self._capacity:
            self._cache.popitem(last=False)

    def __len__(self) -> int:
        return len(self._cache)

    def clear(self) -> None:
        self._cache.clear()



import logging
from abc import abstractmethod, ABC
from typing import List, Sequence, Optional

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LinearSegmentedColormap
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

from . import BinaryClassificationMetric
from .eval_stats_base import PredictionEvalStats, Metric, EvalStatsCollection, PredictionArray, EvalStatsPlot, Array
from ...util import kwarg_if_not_none
from ...util.plot import HistogramPlot
from ...vector_model import VectorRegressionModel, InputOutputData

log = logging.getLogger(__name__)


class RegressionMetric(Metric["RegressionEvalStats"], ABC):
    def compute_value_for_eval_stats(self, eval_stats: "RegressionEvalStats"):
        weights = np.array(eval_stats.weights) if eval_stats.weights is not None else None
        return self.compute_value(np.array(eval_stats.y_true), np.array(eval_stats.y_predicted),
            model=eval_stats.model,
            io_data=eval_stats.ioData,
            **kwarg_if_not_none("weights", weights))

    @abstractmethod
    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray,
            model: VectorRegressionModel = None,
            io_data: InputOutputData = None,
            weights: Optional[np.ndarray] = None):
        pass

    @classmethod
    def compute_errors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return y_predicted - y_true

    @classmethod
    def compute_abs_errors(cls, y_true: np.ndarray, y_predicted: np.ndarray):
        return np.abs(cls.compute_errors(y_true, y_predicted))


class RegressionMetricMAE(RegressionMetric):
    name = "MAE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return mean_absolute_error(y_true, y_predicted, sample_weight=weights)


class RegressionMetricMSE(RegressionMetric):
    name = "MSE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return mean_squared_error(y_true, y_predicted, sample_weight=weights)


class RegressionMetricRMSE(RegressionMetric):
    name = "RMSE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return np.sqrt(mean_squared_error(y_true, y_predicted, sample_weight=weights))


class RegressionMetricRRSE(RegressionMetric):
    name = "RRSE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        r2 = r2_score(y_true, y_predicted, sample_weight=weights)
        return np.sqrt(1 - r2)


class RegressionMetricR2(RegressionMetric):
    name = "R2"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return r2_score(y_true, y_predicted, sample_weight=weights)


class RegressionMetricPCC(RegressionMetric):
    """
    Pearson's correlation coefficient, aka Pearson's R.
    This metric does not consider sample weights.
    """
    name = "PCC"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        cov = np.cov([y_true, y_predicted])
        return cov[0][1] / np.sqrt(cov[0][0] * cov[1][1])


class RegressionMetricStdDevAE(RegressionMetric):
    """
    The standard deviation of the absolute error.
    This metric does not consider sample weights.
    """

    name = "StdDevAE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return np.std(self.compute_abs_errors(y_true, y_predicted))


class RegressionMetricMedianAE(RegressionMetric):
    """
    The median absolute error.
    This metric does not consider sample weights.
    """
    name = "MedianAE"

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        return np.median(self.compute_abs_errors(y_true, y_predicted))


class RegressionMetricFromBinaryClassificationMetric(RegressionMetric):
    """
    Supports the computation of binary classification metrics by converting predicted/target values to class labels.
    This metric does not consider sample weights.
    """

    class ClassGenerator(ABC):
        @abstractmethod
        def compute_class(self, predicted_value: float) -> bool:
            """
            Computes the class from the given value

            :param predicted_value: the value predicted by the regressor or regressor target value
            :return: the class
            """
            pass

        @abstractmethod
        def get_metric_qualifier(self) -> str:
            """
            :return: A (short) string which will be added to the original classification metric's name to
                represent the class conversion logic
            """
            pass

    class ClassGeneratorPositiveBeyond(ClassGenerator):
        def __init__(self, min_value_for_positive: float):
            self.min_value_for_positive = min_value_for_positive

        def compute_class(self, predicted_value: float) -> bool:
            return predicted_value >= self.min_value_for_positive

        def get_metric_qualifier(self) -> str:
            return f">={self.min_value_for_positive}"

    def __init__(self, classification_metric: BinaryClassificationMetric,
            class_generator: ClassGenerator):
        """
        :param classification_metric: the classification metric (which shall consider `True` as the positive label)
        :param class_generator: the class generator, which generates `True` and `False` labels from regression values
        """
        super().__init__(name=classification_metric.name + f"[{class_generator.get_metric_qualifier()}]",
            bounds=classification_metric.bounds)
        self.classification_metric = classification_metric
        self.class_generator = class_generator

    def _apply_class_generator(self, y: np.ndarray) -> np.ndarray:
        return np.array([self.class_generator.compute_class(v) for v in y])

    def compute_value(self, y_true: np.ndarray, y_predicted: np.ndarray, model: VectorRegressionModel = None,
            io_data: InputOutputData = None, weights: Optional[np.ndarray] = None):
        y_true = self._apply_class_generator(y_true)
        y_predicted = self._apply_class_generator(y_predicted)
        return self.classification_metric.compute_value(y_true=y_true, y_predicted=y_predicted)


class HeatMapColorMapFactory(ABC):
    @abstractmethod
    def create_color_map(self, min_sample_weight: float, total_weight: float, num_quantization_levels: int):
        pass


class HeatMapColorMapFactoryWhiteToRed(HeatMapColorMapFactory):
    def create_color_map(self, min_sample_weight: float, total_weight: float, num_quantization_levels: int):
        color_nothing = (1, 1, 1)  # white
        color_min_sample = (1, 0.96, 0.96)  # very slightly red
        color_everything = (0.7, 0, 0)  # dark red
        return LinearSegmentedColormap.from_list("whiteToRed",
            ((0, color_nothing), (min_sample_weight/total_weight, color_min_sample), (1, color_everything)),
            num_quantization_levels)


DEFAULT_REGRESSION_METRICS = (RegressionMetricRRSE(), RegressionMetricR2(), RegressionMetricMAE(),
        RegressionMetricMSE(), RegressionMetricRMSE(), RegressionMetricStdDevAE())


class RegressionEvalStats(PredictionEvalStats["RegressionMetric"]):
    """
    Collects data for the evaluation of predicted continuous values and computes corresponding metrics
    """

    # class members controlling plot appearance, which can be centrally overridden by a user if necessary
    HEATMAP_COLORMAP_FACTORY = HeatMapColorMapFactoryWhiteToRed()
    HEATMAP_DIAGONAL_COLOR = "green"
    HEATMAP_ERROR_BOUNDARY_VALUE = None
    HEATMAP_ERROR_BOUNDARY_COLOR = (0.8, 0.8, 0.8)
    SCATTER_PLOT_POINT_COLOR = (0, 0, 1, 0.05)

    def __init__(self, y_predicted: Optional[PredictionArray] = None, y_true: Optional[PredictionArray] = None,
            metrics: Optional[Sequence["RegressionMetric"]] = None, additional_metrics: Sequence["RegressionMetric"] = None,
            model: VectorRegressionModel = None,
            io_data: InputOutputData = None,
            weights: Optional[Array] = None):
        """
        :param y_predicted: the predicted values
        :param y_true: the true values
        :param metrics: the metrics to compute for evaluation; if None, will use DEFAULT_REGRESSION_METRICS
        :param additional_metrics: the metrics to additionally compute
        :param weights: optional data point weights
        """
        self.model = model
        self.ioData = io_data

        if metrics is None:
            metrics = DEFAULT_REGRESSION_METRICS
        metrics = list(metrics)

        super().__init__(y_predicted, y_true, metrics, additional_metrics=additional_metrics, weights=weights)

    def compute_metric_value(self, metric: RegressionMetric) -> float:
        return metric.compute_value_for_eval_stats(self)

    def compute_mse(self):
        """Computes the mean squared error (MSE)"""
        return self.compute_metric_value(RegressionMetricMSE())

    def compute_rrse(self):
        """Computes the root relative squared error"""
        return self.compute_metric_value(RegressionMetricRRSE())

    def compute_pcc(self):
        """Gets the Pearson correlation coefficient (PCC)"""
        return self.compute_metric_value(RegressionMetricPCC())

    def compute_r2(self):
        """Gets the R^2 score"""
        return self.compute_metric_value(RegressionMetricR2())

    def compute_mae(self):
        """Gets the mean absolute error"""
        return self.compute_metric_value(RegressionMetricMAE())

    def compute_rmse(self):
        """Gets the root mean squared error"""
        return self.compute_metric_value(RegressionMetricRMSE())

    def compute_std_dev_ae(self):
        """Gets the standard deviation of the absolute error"""
        return self.compute_metric_value(RegressionMetricStdDevAE())

    def create_eval_stats_collection(self) -> "RegressionEvalStatsCollection":
        """
        For the case where we collected data on multiple dimensions, obtain a stats collection where
        each object in the collection holds stats on just one dimension
        """
        if self.y_true_multidim is None:
            raise Exception("No multi-dimensional data was collected")
        dim = len(self.y_true_multidim)
        stats_list = []
        for i in range(dim):
            stats = RegressionEvalStats(self.y_predicted_multidim[i], self.y_true_multidim[i])
            stats_list.append(stats)
        return RegressionEvalStatsCollection(stats_list)

    def plot_error_distribution(self, bins="auto", title_add=None) -> Optional[plt.Figure]:
        """
        :param bins: bin specification (see :class:`HistogramPlot`)
        :param title_add: a string to add to the title (on a second line)

        :return: the resulting figure object or None
        """
        errors = np.array(self.y_predicted) - np.array(self.y_true)
        title = "Prediction Error Distribution"
        if title_add is not None:
            title += "\n" + title_add
        if bins == "auto" and len(errors) < 100:
            bins = 10  # seaborn can crash with low number of data points and bins="auto" (tries to allocate vast amounts of memory)
        plot = HistogramPlot(errors, bins=bins, kde=True)
        plot.title(title)
        plot.xlabel("error (prediction - ground truth)")
        plot.ylabel("probability density")
        return plot.fig

    def plot_scatter_ground_truth_predictions(self, figure=True, title_add=None, **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to plot in a separate figure and return that figure
        :param title_add: a string to be added to the title in a second line
        :param kwargs: parameters to be passed on to plt.scatter()

        :return:  the resulting figure object or None
        """
        fig = None
        title = "Scatter Plot of Predicted Values vs. Ground Truth"
        if title_add is not None:
            title += "\n" + title_add
        if figure:
            fig = plt.figure(title.replace("\n", " "))
        y_range = [min(self.y_true), max(self.y_true)]
        plt.scatter(self.y_true, self.y_predicted, c=[self.SCATTER_PLOT_POINT_COLOR], zorder=2, **kwargs)
        plt.plot(y_range, y_range, '-', lw=1, label="_not in legend", color="green", zorder=1)
        plt.xlabel("ground truth")
        plt.ylabel("prediction")
        plt.title(title)
        return fig

    def plot_heatmap_ground_truth_predictions(self, figure=True, cmap=None, bins=60, title_add=None, error_boundary: Optional[float] = None,
            weighted: bool = False, ax: Optional[plt.Axes] = None,
            **kwargs) -> Optional[plt.Figure]:
        """
        :param figure: whether to create a new figure and return that figure (only applies if ax is None)
        :param cmap: the colour map to use (see corresponding parameter of plt.imshow for further information); if None, use factory
            defined in HEATMAP_COLORMAP_FACTORY (which can be centrally set to achieve custom behaviour throughout an application)
        :param bins: how many bins to use for constructing the heatmap
        :param title_add: a string to add to the title (on a second line)
        :param error_boundary: if not None, add two lines (above and below the diagonal) indicating this absolute regression error boundary;
            if None (default), use static member HEATMAP_ERROR_BOUNDARY_VALUE (which is also None by default, but can be centrally set
            to achieve custom behaviour throughout an application)
        :param weighted: whether to consider data point weights
        :param ax: the axis to plot in. If None, use the current axes (which will be the axis of the newly created figure if figure=True).
        :param kwargs: will be passed to plt.imshow()

        :return:  the newly created figure object (if figure=True) or None
        """
        fig = None
        title = "Heat Map of Predicted Values vs. Ground Truth"
        if title_add:
            title += "\n" + title_add
        if figure and ax is None:
            fig = plt.figure(title.replace("\n", " "))
        if ax is None:
            ax = plt.gca()

        y_range = [min(min(self.y_true), min(self.y_predicted)), max(max(self.y_true), max(self.y_predicted))]

        # diagonal
        ax.plot(y_range, y_range, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_DIAGONAL_COLOR, zorder=2)

        # error boundaries
        if error_boundary is None:
            error_boundary = self.HEATMAP_ERROR_BOUNDARY_VALUE
        if error_boundary is not None:
            d = np.array(y_range)
            offs = np.array([error_boundary, error_boundary])
            ax.plot(d, d + offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)
            ax.plot(d, d - offs, '-', lw=0.75, label="_not in legend", color=self.HEATMAP_ERROR_BOUNDARY_COLOR, zorder=2)

        # heat map
        weights = None if not weighted else self.weights
        heatmap, _, _ = np.histogram2d(self.y_true, self.y_predicted, range=(y_range, y_range), bins=bins, density=False, weights=weights)
        extent = (y_range[0], y_range[1], y_range[0], y_range[1])
        if cmap is None:
            num_quantization_levels = min(1000, len(self.y_predicted))
            if not weighted:
                min_sample_weight = 1.0
                total_weight = len(self.y_predicted)
            else:
                min_sample_weight = np.min(self.weights)
                total_weight = np.sum(self.weights)
            cmap = self.HEATMAP_COLORMAP_FACTORY.create_color_map(min_sample_weight, total_weight, num_quantization_levels)
        ax.imshow(heatmap.T, extent=extent, origin='lower', interpolation="none", cmap=cmap, zorder=1, **kwargs)

        ax.set_xlabel("ground truth")
        ax.set_ylabel("prediction")
        ax.set_title(title)
        return fig


class RegressionEvalStatsCollection(EvalStatsCollection[RegressionEvalStats, RegressionMetric]):
    def __init__(self, eval_stats_list: List[RegressionEvalStats]):
        super().__init__(eval_stats_list)
        self.globalStats = None

    def get_combined_eval_stats(self) -> RegressionEvalStats:
        if self.globalStats is None:
            y_true = np.concatenate([evalStats.y_true for evalStats in self.statsList])
            y_predicted = np.concatenate([evalStats.y_predicted for evalStats in self.statsList])
            es0 = self.statsList[0]
            self.globalStats = RegressionEvalStats(y_predicted, y_true, metrics=es0.metrics)
        return self.globalStats


class RegressionEvalStatsPlot(EvalStatsPlot[RegressionEvalStats], ABC):
    pass


class RegressionEvalStatsPlotErrorDistribution(RegressionEvalStatsPlot):
    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_error_distribution(title_add=subtitle)


class RegressionEvalStatsPlotHeatmapGroundTruthPredictions(RegressionEvalStatsPlot):
    def __init__(self, weighted: bool = False):
        self.weighted = weighted

    def is_applicable(self, eval_stats: RegressionEvalStats) -> bool:
        if self.weighted:
            return eval_stats.weights is not None
        else:
            return True

    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_heatmap_ground_truth_predictions(title_add=subtitle, weighted=self.weighted)


class RegressionEvalStatsPlotScatterGroundTruthPredictions(RegressionEvalStatsPlot):
    def create_figure(self, eval_stats: RegressionEvalStats, subtitle: str) -> plt.Figure:
        return eval_stats.plot_scatter_ground_truth_predictions(title_add=subtitle)



"""
Local coordinate systems (for geographic data)
"""
import math
from functools import reduce
from typing import Tuple, Union, List

import numpy as np
import utm
from shapely.geometry import polygon, multipolygon, point, LineString, mapping
from shapely.ops import polygonize, unary_union


class LocalCoordinateSystem(object):
    """
    Represents a local coordinate system for the conversion of geo-coordinates
    (latitude, longitude) to a local Cartesian coordinate system (unit=metre) and vice versa
    using the UTM transform
    """

    def __init__(self, lat, lon):
        """
        Parameters:
            lat: the latitude of the origin of the coordinate system
            lon: the longitude of the origin of the coordinate system
        """
        self.uRef = utm.from_latlon(lat, lon)
        self.uRefE = self.uRef[0]
        self.uRefN = self.uRef[1]
        self.uRefPseudoN = self._pseudo_northing(self.uRefN)

    def get_local_coords(self, lat, lon) -> Tuple[float, float]:
        uE, uN, zM, zL = utm.from_latlon(lat, lon)
        x = uE - self.uRefE
        y = self._pseudo_northing(uN) - self.uRefPseudoN
        return x, y

    def get_lat_lon(self, local_x, local_y) -> Tuple[float, float]:
        easting = local_x + self.uRefE
        pseudo_northing = local_y + self.uRefPseudoN
        return utm.to_latlon(easting, self._real_northing(pseudo_northing), self.uRef[2], self.uRef[3])

    @staticmethod
    def _pseudo_northing(real_northing):
        if real_northing >= 10000000:
            return real_northing - 10000000
        else:
            return real_northing

    @staticmethod
    def _real_northing(pseudo_northing):
        if pseudo_northing < 0:
            return pseudo_northing + 10000000
        else:
            return pseudo_northing


class LocalHexagonalGrid:
    """
    A local hexagonal grid, where hex cells can be referenced by two integer coordinates relative to
    the central grid cell, whose centre is at local coordinate (0, 0) and where positive x-coordinates/columns
    are towards the east and positive y-coordinates/rows are towards the north.
    Every odd row of cells is shifted half a hexagon to the right, i.e. column x for row 1 is half a grid cell
    further to the right than column x for row 0.

    For visualisation purposes, see https://www.redblobgames.com/grids/hexagons/
    """
    def __init__(self, radius_m):
        """
        :param radius_m: the radius, in metres, of each hex cell
        """
        self.radius_m = radius_m
        start_angle = math.pi / 6
        step_angle = math.pi / 3
        self.offset_vectors = []
        for i in range(6):
            angle = start_angle + i * step_angle
            x = math.cos(angle) * radius_m
            y = math.sin(angle) * radius_m
            self.offset_vectors.append(np.array([x, y]))
        self.hexagon_width = 2 * self.offset_vectors[0][0]
        self.hexagon_height = 2 * self.offset_vectors[1][1]
        self.row_step = 0.75 * self.hexagon_height
        self.polygon_area = 6 * self.hexagon_height * self.hexagon_width / 8

    def get_hexagon(self, x_column: int, y_row: int) -> polygon.Polygon:
        """
        Gets the hexagon (polygon) for the given integer hex cell coordinates
        :param x_column: the column coordinate
        :param y_row: the row coordinate
        :return: the hexagon
        """
        centre_x = x_column * self.hexagon_width
        centre_y = y_row * self.row_step
        if y_row % 2 == 1:
            centre_x += 0.5 * self.hexagon_width
        centre = np.array([centre_x, centre_y])
        return polygon.Polygon([centre + o for o in self.offset_vectors])

    def get_min_hexagon_column(self, x):
        lowest_x_definitely_in_column0 = 0
        return math.floor((x - lowest_x_definitely_in_column0) / self.hexagon_width)

    def get_max_hexagon_column(self, x):
        highest_x_definitely_in_column0 = self.hexagon_width / 2
        return math.ceil((x - highest_x_definitely_in_column0) / self.hexagon_width)

    def get_min_hexagon_row(self, y):
        lowest_y_definitely_in_row0 = -self.hexagon_height / 4
        return math.floor((y - lowest_y_definitely_in_row0) / self.row_step)

    def get_max_hexagon_row(self, y):
        highest_y_definitely_in_row0 = self.hexagon_height / 4
        return math.ceil((y - highest_y_definitely_in_row0) / self.row_step)

    def get_hexagon_coord_span_for_bounding_box(self, min_x, min_y, max_x, max_y) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """
        Gets the range of hex-cell coordinates that cover the given bounding box

        :param min_x: minimum x-coordinate of bounding box
        :param min_y: minimum y-coordinate of bounding box
        :param max_x: maximum  x-coordinate of bounding box
        :param max_y: maximum y-coordinate of bounding box
        :return: a pair of pairs ((minCol, min_row), (maxCol, max_row)) indicating the span of cell coordinates
        """
        if min_x > max_x or min_y > max_y:
            raise ValueError()
        min_column = self.get_min_hexagon_column(min_x)
        max_column = self.get_max_hexagon_column(max_x)
        min_row = self.get_min_hexagon_row(min_y)
        max_row = self.get_max_hexagon_row(max_y)
        return ((min_column, min_row), (max_column, max_row))

    def get_hexagon_coords_for_point(self, x, y):
        ((minColumn, minRow), (maxColumn, maxRow)) = self.get_hexagon_coord_span_for_bounding_box(x, y, x, y)
        for xCol in range(minColumn, maxColumn+1):
            for yRow in range(minRow, maxRow+1):
                if self.get_hexagon(xCol, yRow).contains(point.Point(x, y)):
                    return xCol, yRow
        raise Exception("No Hexagon matched; possible edge case (point on hexagon boundary)")


def fix_polygon(poly: Union[polygon.Polygon, multipolygon.MultiPolygon], maxAreaDiff=1e-2) \
        -> Union[polygon.Polygon, multipolygon.MultiPolygon]:
    """
    Fix invalid shapely polygons or multipolygons.

    Reference:
    https://stackoverflow.com/questions/35110632/splitting-self-intersecting-polygon-only-returned-one-polygon-in-shapely

    :param poly: the polygon to fix
    :param maxAreaDiff: the maximum change in area
    :return: the fixed polygon or None if it cannot be fixed given the area change constraint
    """
    def _fix_polygon_component(coords: List[Tuple[float, float]]):
        res = list(polygonize(unary_union(LineString(list(coords) + [coords[0]]))))
        return reduce(lambda p1, p2: p1.union(p2), res)

    if poly.is_valid:
        return poly
    else:
        if isinstance(poly, polygon.Polygon):
            exterior_coords = poly.exterior.coords[:]
            fixed_exterior = _fix_polygon_component(exterior_coords)
            fixed_interior = polygon.Polygon()
            for interior in poly.interiors:
                coords = interior.coords[:]
                fixed_interior = fixed_interior.union(_fix_polygon_component(coords))
            fixed_polygon = fixed_exterior.difference(fixed_interior)
        elif isinstance(poly, multipolygon.MultiPolygon):
            polys = list(poly)
            fixed_polys = [fix_polygon(p, maxAreaDiff=maxAreaDiff) for p in polys]
            fixed_polygon = reduce(lambda p1, p2: p1.union(p2), fixed_polys)
        else:
            raise Exception(f"Unsupported type {type(poly)}")
        area_diff = float('Inf') if poly.area == 0 else abs(poly.area - fixed_polygon.area) / poly.area
        #log.info(f"Invalid polygon\n{poly}\nComputed fix:\n{fixed_polygon}.\nArea error: {area_diff}")
        if area_diff > maxAreaDiff:
            return None
        else:
            return fixed_polygon




import collections
from math import log, exp
from typing import Optional

import numpy as np
import pandas as pd

from .vector_model import VectorClassificationModel


class CategoricalNaiveBayesVectorClassificationModel(VectorClassificationModel):
    """
    Naive Bayes with categorical features
    """
    def __init__(self, pseudo_count=0.1):
        """
        :param pseudo_count: the count to add to each empirical count in order to avoid overfitting
        """
        super().__init__()
        self.prior = None
        self.conditionals = None
        self.pseudoCount = pseudo_count

    def _fit_classifier(self, x: pd.DataFrame, y: pd.DataFrame, weights: Optional[pd.Series] = None):
        self._warn_sample_weights_unsupported(False, weights)
        self.prior = collections.defaultdict(lambda: 0)
        self.conditionals = collections.defaultdict(lambda: [collections.defaultdict(lambda: 0) for _ in range(x.shape[1])])
        increment = 1
        for idxRow in range(x.shape[0]):
            cls = y.iloc[idxRow, 0]
            self.prior[cls] += increment
            for idxFeature in range(x.shape[1]):
                value = x.iloc[idxRow, idxFeature]
                self.conditionals[cls][idxFeature][value] += increment
        # get rid of defaultdicts, which are not picklable
        self.prior = dict(self.prior)
        self.conditionals = {k: [dict(d) for d in l] for k, l in self.conditionals.items()}

    def _predict_class_probabilities(self, x: pd.DataFrame):
        results = []
        for _, features in x.iterrows():
            class_probabilities = np.zeros(len(self._labels))
            for i, cls in enumerate(self._labels):
                lp = log(self._probability(self.prior, cls))
                for idx_feature, value in enumerate(features):
                    lp += log(self._probability(self.conditionals[cls][idx_feature], value))
                class_probabilities[i] = exp(lp)
            class_probabilities /= np.sum(class_probabilities)
            results.append(class_probabilities)
        return pd.DataFrame(results, columns=self._labels)

    def _probability(self, counts, value):
        value_count = counts.get(value, 0.0)
        total_count = sum(counts.values())
        return (value_count + self.pseudoCount) / (total_count + self.pseudoCount)

    def _predict(self, x: pd.DataFrame) -> pd.DataFrame:
        results = []
        for _, features in x.iterrows():
            best_cls = None
            best_lp = None
            for cls in self.prior:
                lp = log(self._probability(self.prior, cls))
                for idxFeature, value in enumerate(features):
                    lp += log(self._probability(self.conditionals[cls][idxFeature], value))
                if best_lp is None or lp > best_lp:
                    best_lp = lp
                    best_cls = cls
            results.append(best_cls)
        return pd.DataFrame(results, columns=self.get_predicted_variable_names())



import logging
from typing import Optional, List, Union

import pandas as pd
import torch

from ....torch import TorchVectorRegressionModel, Tensoriser, NNOptimiserParams, TorchModel
from ....torch.torch_models.seq.seq_modules import EncoderDecoderModule, EncoderFactory, DecoderFactory
from ....vectoriser import SequenceVectoriser

log = logging.getLogger(__name__)


class EncoderDecoderVectorRegressionModel(TorchVectorRegressionModel):
    """
    A highly general encoder-decoder sequence model, which encodes a sequence of history items
    and uses the encoding to make predictions for one or more target sequence items.
    History and target sequences are converted to vectors via vectorisers.
    """
    def __init__(self, cuda: bool,
            history_sequence_column_name: str,
            history_sequence_vectoriser: SequenceVectoriser,
            history_sequence_variable_length: bool,
            target_sequence_column_name: str,
            target_sequence_vectoriser: SequenceVectoriser,
            latent_dim: int,
            encoder_factory: EncoderFactory,
            decoder_factory: DecoderFactory,
            nn_optimiser_params: Optional[NNOptimiserParams] = None):
        """
        :param cuda: whether to use a CUDA device
        :param history_sequence_column_name: the name of the data frame input column which contains the history sequences to be encoded.
            The column must contain a sequence of items that can be converted to vectors via the `history_sequence_vectorizer`
        :param history_sequence_vectoriser: a vectorizer which converts history sequence items to vectors
        :param history_sequence_variable_length: whether history sequences can be of variable length
        :param target_sequence_column_name: the column containing the target item sequence; Note that the column must
            contain sequences even if there is but a single target item for which predictions shall be made.
            In such cases, simply use a column that contains lists with a single item each.
        :param target_sequence_vectoriser: the vectoriser for the generation of feature vectors for the target
            items.
        :param latent_dim: the number of latent dimensions to be used by the encoder
        :param encoder_factory: a factory for the creation of the encoder, which takes sequence items from the history
            and encodes them into vectors of dimension `latent_dim`
        :param decoder_factory: a factory for the creation of the decoder component, which takes a latent vector produced by the
            encoder and (a sequence of) target features to make predictions
        :param nn_optimiser_params: the optimiser parameters
        """
        super().__init__(self._create_model, nn_optimiser_params=nn_optimiser_params)
        self.history_sequence_variable_length = history_sequence_variable_length
        self.latent_dim = latent_dim
        self.cuda = cuda
        self.decoder_factory = decoder_factory
        self.encoder_factory = encoder_factory
        self.sequenceVectoriser = history_sequence_vectoriser
        self.history_sequence_dim_per_item: Optional[int] = None
        self.target_feature_dim: Optional[int] = None
        self.with_input_tensoriser(self.InputTensoriser(history_sequence_column_name, history_sequence_vectoriser,
            target_sequence_column_name, target_sequence_vectoriser))

    def _create_model(self):
        return self.EncoderDecoderModel(self)

    class InputTensoriser(Tensoriser):
        def __init__(self,
                history_sequence_column_name: str,
                history_sequence_vectoriser: SequenceVectoriser,
                target_sequence_column_name: str,
                target_sequence_vectoriser: SequenceVectoriser):
            self.history_sequence_column_name = history_sequence_column_name
            self.history_sequence_vectoriser = history_sequence_vectoriser
            self.target_sequence_column_name = target_sequence_column_name
            self.target_sequence_vectoriser = target_sequence_vectoriser

        def fit(self, df: pd.DataFrame, model=None):
            model: "EncoderDecoderVectorRegressionModel"
            self.history_sequence_vectoriser.fit(df[self.history_sequence_column_name])
            model.history_sequence_dim_per_item = self.history_sequence_vectoriser.get_vector_dim(df[self.history_sequence_column_name].iloc[0])
            self.target_sequence_vectoriser.fit(df[self.target_sequence_column_name])
            model.target_feature_dim = self.target_sequence_vectoriser.get_vector_dim(df[self.target_sequence_column_name].iloc[0])

        def _tensorise(self, df: pd.DataFrame) -> Union[torch.Tensor, List[torch.Tensor]]:
            log.debug(f"Applying {self} to data frame of length {len(df)} ...")
            history = df[self.history_sequence_column_name]
            history_sequences, history_sequence_lengths = self.history_sequence_vectoriser.apply_multi_with_padding(history)
            targets = df[self.target_sequence_column_name]
            target_sequences, target_sequence_lengths = self.target_sequence_vectoriser.apply_multi_with_padding(targets)
            return [
                torch.tensor(history_sequences).float(),
                torch.tensor(history_sequence_lengths),
                torch.tensor(target_sequences).float(),
                torch.tensor(target_sequence_lengths),
            ]

    class EncoderDecoderModel(TorchModel):
        def __init__(self, parent: "EncoderDecoderVectorRegressionModel"):
            super().__init__(parent.cuda)
            self.parent = parent

        def create_torch_module(self) -> torch.nn.Module:
            latent_dim = self.parent.latent_dim
            target_feature_dim = self.parent.target_feature_dim
            encoder = self.parent.encoder_factory.create_encoder(self.parent.history_sequence_dim_per_item, latent_dim)
            decoder = self.parent.decoder_factory.create_decoder(latent_dim, target_feature_dim)
            return EncoderDecoderModule(encoder, decoder, self.parent.history_sequence_variable_length)



from typing import Callable, Optional, Sequence

import torch
from torch import nn

from ...torch_base import MCDropoutCapableNNModule
from ....util.string import object_repr, function_name


class MultiLayerPerceptron(MCDropoutCapableNNModule):
    def __init__(self, input_dim: float, output_dim: float, hidden_dims: Sequence[int],
            hid_activation_fn: Callable[[torch.Tensor], torch.Tensor] = torch.sigmoid,
            output_activation_fn: Optional[Callable[[torch.Tensor], torch.Tensor]] = torch.sigmoid,
            p_dropout: Optional[float] = None):
        super().__init__()
        self.inputDim = input_dim
        self.outputDim = output_dim
        self.hiddenDims = hidden_dims
        self.hidActivationFn = hid_activation_fn
        self.outputActivationFn = output_activation_fn
        self.pDropout = p_dropout
        self.layers = nn.ModuleList()
        if p_dropout is not None:
            self.dropout = nn.Dropout(p=p_dropout)
        else:
            self.dropout = None
        prev_dim = input_dim
        for dim in [*hidden_dims, output_dim]:
            self.layers.append(nn.Linear(prev_dim, dim))
            prev_dim = dim

    def __str__(self):
        return object_repr(self, dict(inputDim=self.inputDim, outputDim=self.outputDim, hiddenDims=self.hiddenDims,
            hidActivationFn=function_name(self.hidActivationFn) if self.hidActivationFn is not None else None,
            outputActivationFn=function_name(self.outputActivationFn) if self.outputActivationFn is not None else None,
            pDropout=self.pDropout))

    def forward(self, x):
        for i, layer in enumerate(self.layers):
            is_last = i+1 == len(self.layers)
            x = layer(x)
            if not is_last and self.dropout is not None:
                x = self.dropout(x)
            activation = self.hidActivationFn if not is_last else self.outputActivationFn
            if activation is not None:
                x = activation(x)
        return x



import geopandas as gp
import numpy as np
from abc import ABC, abstractmethod
from shapely.geometry import MultiPoint
from typing import Union

from ...clustering import EuclideanClusterer

TCoordinates = Union[np.ndarray, MultiPoint, gp.GeoDataFrame, EuclideanClusterer.Cluster]


def validate_coordinates(coordinates: np.ndarray):
    # for the moment we only support 2-dim coordinates. We can adjust it in the future when needed
    if not len(coordinates.shape) == 2 or coordinates.shape[1] != 2:
        raise Exception(f"Coordinates must be of shape (n, 2), instead got: {coordinates.shape}")


def extract_coordinates_array(coordinates: TCoordinates) -> np.ndarray:
    """
    Extract coordinates as numpy array
    """
    if isinstance(coordinates, gp.GeoDataFrame):
        try:
            coordinates = np.array(MultiPoint(list(coordinates.geometry)))
        except Exception:
            raise ValueError(f"Could not extract coordinates from GeoDataFrame. "
                             f"Is the geometry column a sequence of Points?")
    elif isinstance(coordinates, MultiPoint):
        coordinates = np.array(coordinates)
    elif isinstance(coordinates, EuclideanClusterer.Cluster):
        coordinates = coordinates.datapoints
    validate_coordinates(coordinates)
    return coordinates


class GeoDataFrameWrapper(ABC):
    @abstractmethod
    def to_geodf(self, *args, **kwargs) -> gp.GeoDataFrame:
        pass

    def plot(self, *args, **kwargs):
        self.to_geodf().plot(*args, **kwargs)



from tensorflow import keras

from .tf_base import KerasVectorRegressionModel
from .. import normalisation


class KerasMultiLayerPerceptronVectorRegressionModel(KerasVectorRegressionModel):
    def __init__(self, hidden_dims=(5,5), hidden_activation="sigmoid", output_activation="sigmoid", loss="mse",
            metrics=("mse",), optimiser="adam", normalisation_mode=normalisation.NormalisationMode.MAX_BY_COLUMN, **kwargs):
        super().__init__(normalisation_mode, loss, metrics, optimiser, **kwargs)
        self.hiddenDims = hidden_dims
        self.hiddenActivation = hidden_activation
        self.outputActivation = output_activation

    def __str__(self):
        params = dict(hiddenDims=self.hiddenDims, hiddenActivation=self.hiddenActivation,
            outputActivation=self.outputActivation)
        return f"{self.__class__.__name__}{params}={super().__str__()}"

    def _create_model(self, input_dim, output_dim):
        model_inputs = keras.Input(shape=(input_dim,), name='input')
        x = model_inputs
        for i, hiddenDim in enumerate(self.hiddenDims):
            x = keras.layers.Dense(hiddenDim, activation=self.hiddenActivation, name='dense_%d' % i)(x)
        model_outputs = keras.layers.Dense(output_dim, activation=self.outputActivation, name='predictions')(x)
        return keras.Model(inputs=model_inputs, outputs=model_outputs)





import collections
import copy
import logging
import re
from abc import ABC, abstractmethod
from typing import Dict, Union, Sequence, List, Tuple, Optional

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from .data import InputOutputData
from .evaluation.crossval import VectorModelCrossValidationData
from .util.deprecation import deprecated
from .util.plot import MATPLOTLIB_DEFAULT_FIGURE_SIZE
from .util.string import ToStringMixin
from .vector_model import VectorModel

log = logging.getLogger(__name__)


class FeatureImportance:
    def __init__(self, feature_importance_dict: Union[Dict[str, float], Dict[str, Dict[str, float]]]):
        self.feature_importance_dict = feature_importance_dict
        self._isMultiVar = self._is_dict(next(iter(feature_importance_dict.values())))

    @staticmethod
    def _is_dict(x):
        return hasattr(x, "get")

    def get_feature_importance_dict(self, predicted_var_name=None) -> Dict[str, float]:
        if self._isMultiVar:
            self.feature_importance_dict: Dict[str, Dict[str, float]]
            if predicted_var_name is not None:
                return self.feature_importance_dict[predicted_var_name]
            else:
                if len(self.feature_importance_dict) > 1:
                    raise ValueError("Must provide predicted variable name (multiple output variables)")
                else:
                    return next(iter(self.feature_importance_dict.values()))
        else:
            return self.feature_importance_dict

    def get_sorted_tuples(self, predicted_var_name=None, reverse=False) -> List[Tuple[str, float]]:
        """
        :param predicted_var_name: the predicted variable name for which to retrieve the sorted feature importance values
        :param reverse: whether to reverse the order (i.e. descending order of importance values, where the most important feature comes
            first, rather than ascending order)
        :return: a sorted list of tuples (feature name, feature importance)
        """
        # noinspection PyTypeChecker
        tuples: List[Tuple[str, float]] = list(self.get_feature_importance_dict(predicted_var_name).items())
        tuples.sort(key=lambda t: t[1], reverse=reverse)
        return tuples

    def plot(self, predicted_var_name=None, sort=True) -> plt.Figure:
        return plot_feature_importance(self.get_feature_importance_dict(predicted_var_name=predicted_var_name), sort=sort)

    def get_data_frame(self, predicted_var_name=None) -> pd.DataFrame:
        """
        :param predicted_var_name: the predicted variable name
        :return: a data frame with two columns, "feature" and "importance"
        """
        names_and_importance = self.get_sorted_tuples(predicted_var_name=predicted_var_name, reverse=True)
        return pd.DataFrame(names_and_importance, columns=["feature", "importance"])


class FeatureImportanceProvider(ABC):
    """
    Interface for models that can provide feature importance values
    """
    @abstractmethod
    def get_feature_importance_dict(self) -> Union[Dict[str, float], Dict[str, Dict[str, float]]]:
        """
        Gets the feature importance values

        :return: either a dictionary mapping feature names to importance values or (for models predicting multiple
            variables (independently)) a dictionary which maps predicted variable names to such dictionaries
        """
        pass

    def get_feature_importance(self) -> FeatureImportance:
        return FeatureImportance(self.get_feature_importance_dict())

    @deprecated("Use getFeatureImportanceDict or the high-level interface getFeatureImportance instead.")
    def get_feature_importances(self) -> Union[Dict[str, float], Dict[str, Dict[str, float]]]:
        return self.get_feature_importance_dict()


def plot_feature_importance(feature_importance_dict: Dict[str, float], subtitle: str = None, sort=True) -> plt.Figure:
    if sort:
        feature_importance_dict = {k: v for k, v in sorted(feature_importance_dict.items(), key=lambda x: x[1], reverse=True)}
    num_features = len(feature_importance_dict)
    default_width, default_height = MATPLOTLIB_DEFAULT_FIGURE_SIZE
    height = max(default_height, default_height * num_features / 20)
    fig, ax = plt.subplots(figsize=(default_width, height))
    sns.barplot(x=list(feature_importance_dict.values()), y=list(feature_importance_dict.keys()), ax=ax)
    title = "Feature Importance"
    if subtitle is not None:
        title += "\n" + subtitle
    plt.title(title)
    plt.tight_layout()
    return fig


class AggregatedFeatureImportance:
    """
    Aggregates feature importance values (e.g. from models implementing FeatureImportanceProvider, such as sklearn's RandomForest
    models and compatible models from lightgbm, etc.)
    """
    def __init__(self, *items: Union[FeatureImportanceProvider, Dict[str, float], Dict[str, Dict[str, float]]],
            feature_agg_reg_ex: Sequence[str] = (), agg_fn=np.mean):
        r"""
        :param items: (optional) initial list of feature importance providers or dictionaries to aggregate; further
            values can be added via method add
        :param feature_agg_reg_ex: a sequence of regular expressions describing which feature names to sum as one. Each regex must
            contain exactly one group. If a regex matches a feature name, the feature importance will be summed under the key
            of the matched group instead of the full feature name. For example, the regex r"(\w+)_\d+$" will cause "foo_1" and "foo_2"
            to be summed under "foo" and similarly "bar_1" and "bar_2" to be summed under "bar".
        """
        self._agg_dict = None
        self._is_nested = None
        self._num_dicts_added = 0
        self._feature_agg_reg_ex = [re.compile(p) for p in feature_agg_reg_ex]
        self._agg_fn = agg_fn
        for item in items:
            self.add(item)

    @staticmethod
    def _is_dict(x):
        return hasattr(x, "get")

    def add(self, feature_importance: Union[FeatureImportanceProvider, Dict[str, float], Dict[str, Dict[str, float]]]):
        """
        Adds the feature importance values from the given dictionary

        :param feature_importance: the dictionary obtained via a model's getFeatureImportances method
        """
        if isinstance(feature_importance, FeatureImportanceProvider):
            feature_importance = feature_importance.get_feature_importance_dict()
        if self._is_nested is None:
            self._is_nested = self._is_dict(next(iter(feature_importance.values())))
        if self._is_nested:
            if self._agg_dict is None:
                self._agg_dict = collections.defaultdict(lambda: collections.defaultdict(list))
            for targetName, d in feature_importance.items():
                d: dict
                for featureName, value in d.items():
                    self._agg_dict[targetName][self._agg_feature_name(featureName)].append(value)
        else:
            if self._agg_dict is None:
                self._agg_dict = collections.defaultdict(list)
            for featureName, value in feature_importance.items():
                self._agg_dict[self._agg_feature_name(featureName)].append(value)
        self._num_dicts_added += 1

    def _agg_feature_name(self, feature_name: str):
        for regex in self._feature_agg_reg_ex:
            m = regex.match(feature_name)
            if m is not None:
                return m.group(1)
        return feature_name

    def get_aggregated_feature_importance_dict(self) -> Union[Dict[str, float], Dict[str, Dict[str, float]]]:
        def aggregate(d: dict):
            return {k: self._agg_fn(l) for k, l in d.items()}

        if self._is_nested:
            return {k: aggregate(d) for k, d in self._agg_dict.items()}
        else:
            return aggregate(self._agg_dict)

    def get_aggregated_feature_importance(self) -> FeatureImportance:
        return FeatureImportance(self.get_aggregated_feature_importance_dict())


def compute_permutation_feature_importance_dict(model, io_data: InputOutputData, scoring, num_repeats: int, random_state,
        exclude_input_preprocessors=False, num_jobs=None):
    from sklearn.inspection import permutation_importance
    if exclude_input_preprocessors:
        inputs = model.compute_model_inputs(io_data.inputs)
        model = copy.copy(model)
        model.remove_input_preprocessors()
    else:
        inputs = io_data.inputs
    feature_names = inputs.columns
    pi = permutation_importance(model, inputs, io_data.outputs, n_repeats=num_repeats, random_state=random_state, scoring=scoring,
        n_jobs=num_jobs)
    importance_values = pi.importances_mean
    assert len(importance_values) == len(feature_names)
    feature_importance_dict = dict(zip(feature_names, importance_values))
    return feature_importance_dict


class AggregatedPermutationFeatureImportance(ToStringMixin):
    def __init__(self, aggregated_feature_importance: AggregatedFeatureImportance, scoring, num_repeats=5, random_seed=42,
            exclude_model_input_preprocessors=False, num_jobs: Optional[int] = None):
        """
        :param aggregated_feature_importance: the object in which to aggregate the feature importance (to which no feature importance
            values should have yet been added)
        :param scoring: the scoring method; see https://scikit-learn.org/stable/modules/model_evaluation.html; e.g. "r2" for regression or
            "accuracy" for classification
        :param num_repeats: the number of data permutations to apply for each model
        :param random_seed: the random seed for shuffling the data
        :param exclude_model_input_preprocessors: whether to exclude model input preprocessors, such that the
            feature importance will be reported on the transformed inputs that are actually fed to the model rather than the original
            inputs.
            Enabling this can, for example, help save time in cases where the input preprocessors discard many of the raw input
            columns, but it may not be a good idea of the preprocessors generate multiple columns from the original input columns.
        :param num_jobs:
            Number of jobs to run in parallel. Each separate model-data permutation feature importance computation is parallelised over
            the columns. `None` means 1 unless in a :obj:`joblib.parallel_backend` context.
            `-1` means using all processors.
        """
        self._agg = aggregated_feature_importance
        self.scoring = scoring
        self.numRepeats = num_repeats
        self.randomSeed = random_seed
        self.excludeModelInputPreprocessors = exclude_model_input_preprocessors
        self.numJobs = num_jobs

    def add(self, model: VectorModel, io_data: InputOutputData):
        feature_importance_dict = compute_permutation_feature_importance_dict(model, io_data, self.scoring, num_repeats=self.numRepeats,
            random_state=self.randomSeed, exclude_input_preprocessors=self.excludeModelInputPreprocessors, num_jobs=self.numJobs)
        self._agg.add(feature_importance_dict)

    def add_cross_validation_data(self, cross_val_data: VectorModelCrossValidationData):
        if cross_val_data.trained_models is None:
            raise ValueError("No models in cross-validation data; enable model collection during cross-validation")
        for i, (model, evalData) in enumerate(zip(cross_val_data.trained_models, cross_val_data.eval_data_list), start=1):
            log.info(f"Computing permutation feature importance for model #{i}/{len(cross_val_data.trained_models)}")
            self.add(model, evalData.io_data)

    def get_feature_importance(self) -> FeatureImportance:
        return self._agg.get_aggregated_feature_importance()



from abc import ABC, abstractmethod
import logging
import math
import os
import re
import subprocess
import tempfile
import time
from typing import List

import numpy as np


log = logging.getLogger(__name__)


class CostScaler:
    """
    Serves to scale floats and converts them into integers (and vice versa) whilst
    maintaining decimal precision
    """

    def __init__(self, cost_values: List[float], significant_digits: int):
        """
        Parameters:
            cost_values: the sequence of cost values whose precision should be maintained in the int realm
            significant_digits: the number of significant digits that shall at least be maintained
        """
        exp10 = significant_digits - 1 - min([0] + [np.floor(np.log10(v)) for v in cost_values])
        self.scalingFactor = math.pow(10, exp10)

    def scaled_int(self, original_value: float) -> int:
        """Returns the scaled value as an integer"""
        return int(round(original_value * self.scalingFactor))

    def scaled_float(self, original_value: float) -> float:
        return original_value * self.scalingFactor

    def original_value(self, scaled_value: float) -> float:
        """Returns the original unscaled value from a scaled value"""
        return scaled_value / self.scalingFactor

    def __str__(self):
        return "CostScaler[factor=%d]" % self.scalingFactor


class MiniZincProblem(ABC):

    def create_mini_zinc_file(self, f):
        """
        Writes MiniZinc code

        :param f: an OS-level handle to an open file
        """
        os.write(f, bytes(self.get_mini_zinc_code(), 'utf-8'))

    @abstractmethod
    def get_mini_zinc_code(self):
        pass


class MiniZincSolver(object):
    log = log.getChild(__qualname__)

    def __init__(self, name='OSICBC', solver_time_seconds=None, fzn_output_path=None):
        """
        :param name: name of solver compatible with miniZinc
        :param solver_time_seconds: upper time limit for solver in seconds
        :param fzn_output_path: flatZinc output path
        """
        self.solver_name = name
        self.solver_time_limit_secs = solver_time_seconds
        self.fzn_output_path = fzn_output_path
        self.last_solver_time_secs = None
        self.last_solver_output = None
        self.lastSolverErrOutput = None

    def __str(self):
        return f"MiniZincSolver[{self.solver_name}]"

    def solve_path(self, mzn_path: str, log_info=True) -> str:
        """
        Solves the MiniZinc problem stored at the given file path

        :param mzn_path: path to file containing MiniZinc problem code
        :param log_info: whether to log solver output at INFO level rather than DEBUG level
        :return: the solver output
        """
        self.last_solver_time_secs = None
        log_solver = self.log.info if log_info else self.log.debug

        args = ["--statistics", "--solver", self.solver_name]
        if self.solver_time_limit_secs is not None:
            args.append("--time-limit")
            args.append(str(self.solver_time_limit_secs * 1000))
        if self.fzn_output_path is not None:
            args.append("--output-fzn-to-file")
            args.append(self.fzn_output_path)
        args.append(mzn_path)
        command = "minizinc " + " ".join(args)

        self.log.info("Running %s" % command)
        start_time = time.time()
        proc = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        output = []
        while True:
            line = proc.stdout.readline().decode("utf-8")
            if not line:
                break
            output.append(line)
            log_solver("Solver output: %s" % line.rstrip())
        output = "".join(output)
        proc.wait()
        if proc.returncode != 0:
            raise Exception(f"MiniZinc call failed with return code {proc.returncode}; output: {output}")
        self.last_solver_time_secs = time.time() - start_time
        self.last_solver_output = output
        self.log.info("Solver time: %.1fs" % self.last_solver_time_secs)
        return output

    def solve_problem(self, problem: MiniZincProblem, keep_temp_file=False, log_info=True) -> str:
        """
        Solves the given MiniZinc problem

        :param problem: the problem to solve
        :param keep_temp_file: whether to keep the temporary .mzv file
        :param log_info: whether to log solver output at INFO level rather than DEBUG level
        :return: the solver output
        """
        f, path = tempfile.mkstemp(".mzn")
        try:
            try:
                problem.create_mini_zinc_file(f)
            finally:
                os.close(f)
            return self.solve_path(path, log_info=log_info)
        finally:
            if not keep_temp_file:
                os.unlink(path)

    def get_last_solver_time_secs(self):
        return self.last_solver_time_secs


def extract_1d_array_from_output(string_identifier: str, output: str) -> List:
    regexOutput = re.search(r'{stringIdentifier} = array1d\(\d+\.\.\d+, (\[.*?\])'.format(stringIdentifier=string_identifier), output)
    return eval(regexOutput.group(1))


def extract_multi_dim_array_from_output(string_identifier: str, dim: int, output: str, boolean=False) -> np.array:
    dim_regex = r"1..(\d+), "
    regex = r'{stringIdentifier} = array{dim}d\({dimsRegex}(\[.*?\])'.format(stringIdentifier=string_identifier, dim=dim,
        dimsRegex=dim_regex*dim)
    match = re.search(regex, output)
    if match is None:
        raise Exception("No match found for regex: %s" % regex)
    shape = [int(match.group(i)) for i in range(1, dim+1)]
    flat_list = match.group(dim+1)
    if boolean:
        flat_list = flat_list.replace("false", "0").replace("true", "1")
    flat_list = eval(flat_list)
    array1d = np.array(flat_list)
    arraymd = array1d.reshape(shape)
    return arraymd


def array_to_mini_zinc(a: np.array, element_cast):
    shape = a.shape
    dims = ", ".join([f"1..{n}" for n in shape])
    values = str(list(map(element_cast, a.flatten())))
    return f"array{len(shape)}d({dims}, {values})"



import pickle
import hashlib
from typing import Any


def pickle_hash(o: Any, algorithm='sha1', with_class_name=False) -> str:
    """
    Computes a deterministic hash code based on the contents of the given object

    :param o: any picklable object
    :param algorithm: the name of a hashing algorithm supported by hashlib
    :param with_class_name: if True, prepend o's class name to the returned hex string
    :return: the object's hash code as a hex string
    """
    s = pickle.dumps(o)
    result = hashlib.new(algorithm, s).hexdigest()
    if with_class_name:
        result = o.__class__.__name__ + result
    return result


def str_hash(o: Any, algorithm='sha1', with_class_name=False) -> str:
    """
    Computes a deterministic hash code based on the string representation of the given object

    :param o: any object
    :param algorithm: the name of a hashing algorithm supported by hashlib
    :param with_class_name: if True, prepend o's class name to the returned hex string
    :return: the object's hash code as a hex string
    """
    s = str(o)
    result = hashlib.new(algorithm, s).hexdigest()
    if with_class_name:
        result = o.__class__.__name__ + result
    return result


import json

import numpy as np
import pandas as pd


def snapshot_compatible(obj, float_decimals=6, significant_digits=12):
    """
    Renders an object snapshot-compatible by appropriately converting nested types and reducing float precision to a level
    that is likely to not cause problems when testing snapshots for equivalence on different platforms

    :param obj: the object to convert
    :param float_decimals: the number of float decimal places to consider
    :param significant_digits: the (maximum) number of significant digits to consider
    :return: the converted object
    """
    result = json.loads(json.dumps(obj, default=json_mapper))
    return convert_floats(result, float_decimals, significant_digits)


def reduce_float_precision(f, decimals, significant_digits):
    return float(format(float(format(f, '.%df' % decimals)), ".%dg" % significant_digits))


def convert_floats(o, float_decimals, significant_digits):
    if type(o) == list:
        return [convert_floats(x, float_decimals, significant_digits) for x in o]
    elif type(o) == dict:
        return {key: convert_floats(value, float_decimals, significant_digits) for (key, value) in o.items()}
    elif type(o) == float:
        return reduce_float_precision(o, float_decimals, significant_digits)
    else:
        return o


def json_mapper(o):
    """
    Maps the given data object to a representation that is JSON-compatible.
    Currently, the supported object types include, in particular, numpy arrays as well as pandas Series and DataFrames.

    :param o: the object to convert
    :return: the converted object
    """
    if isinstance(o, pd.DataFrame):
        if isinstance(o.index, pd.DatetimeIndex):
            o.index = o.index.astype('int64').tolist()
        return o.to_dict()
    if isinstance(o, pd.Series):
        return o.values.tolist()
    if isinstance(o, np.ndarray):
        return o.tolist()
    if isinstance(o, list):
        return o
    else:
        return o.__dict__



from dataclasses import dataclass
from datetime import datetime
from concurrent.futures import ProcessPoolExecutor
import uuid

import logging
import os
import pandas as pd
from abc import ABC
from abc import abstractmethod
from random import Random
from typing import Dict, Sequence, Any, Callable, Generator, Union, Tuple, List, Optional, Hashable

from .evaluation.evaluator import MetricsDictProvider
from .local_search import SACostValue, SACostValueNumeric, SAOperator, SAState, SimulatedAnnealing, \
    SAProbabilitySchedule, SAProbabilityFunctionLinear
from .tracking.tracking_base import TrackingMixin, TrackedExperiment
from .vector_model import VectorModel

log = logging.getLogger(__name__)


def iter_param_combinations(hyper_param_values: Dict[str, Sequence[Any]]) -> Generator[Dict[str, Any], None, None]:
    """
    Create all possible combinations of values from a dictionary of possible parameter values

    :param hyper_param_values: a mapping from parameter names to lists of possible values
    :return: a dictionary mapping each parameter name to one of the values
    """
    pairs = list(hyper_param_values.items())

    def _iter_recursive_param_combinations(pairs, i, params):
        """
        Recursive function to create all possible combinations from a list of key-array entries.
        :param pairs: a dictionary of parameter names and their corresponding values
        :param i: the recursive step
        :param params: a dictionary for the iteration results
        """
        if i == len(pairs):
            yield dict(params)
        else:
            param_name, param_values = pairs[i]
            for paramValue in param_values:
                params[param_name] = paramValue
                yield from _iter_recursive_param_combinations(pairs, i+1, params)

    return _iter_recursive_param_combinations(pairs, 0, {})


class ParameterCombinationSkipDecider(ABC):
    """
    Abstraction for a functional component which is told all parameter combinations that have been considered
    and can use these as a basis for deciding whether another parameter combination shall be skipped/not be considered.
    """

    @abstractmethod
    def tell(self, params: Dict[str, Any], metrics: Dict[str, Any]):
        """
        Informs the decider about a previously evaluated parameter combination

        :param params: the parameter combination
        :param metrics: the evaluation metrics
        """
        pass

    @abstractmethod
    def is_skipped(self, params: Dict[str, Any]):
        """
        Decides whether the given parameter combination shall be skipped

        :param params:
        :return: True iff it shall be skipped
        """
        pass


class ParameterCombinationEquivalenceClassValueCache(ABC):
    """
    Represents a cache which stores (arbitrary) values for parameter combinations, i.e. keys in the cache
    are derived from parameter combinations.
    The cache may map the equivalent parameter combinations to the same keys to indicate that the
    parameter combinations are equivalent; the keys thus correspond to representations of equivalence classes over
    parameter combinations.
    This enables hyper-parameter search to skip the re-computation of results for equivalent parameter combinations.
    """
    def __init__(self):
        self._cache = {}

    @abstractmethod
    def _equivalence_class(self, params: Dict[str, Any]) -> Hashable:
        """
        Computes a (hashable) equivalence class representation for the given parameter combination.
        For instance, if all parameters have influence on the evaluation of a model and no two combinations would 
        lead to equivalent results, this could simply return a tuple containing all parameter values (in a fixed order).

        :param params: the parameter combination
        :return: a hashable key containing all the information from the parameter combination that influences the
            computation of model evaluation results
        """
        pass

    def set(self, params: Dict[str, Any], value: Any):
        self._cache[self._equivalence_class(params)] = value

    def get(self, params: Dict[str, Any]):
        """
        Gets the value associated with the (equivalence class of the) parameter combination
        :param params: the parameter combination
        :return:
        """
        return self._cache.get(self._equivalence_class(params))


class ParametersMetricsCollection:
    """
    Utility class for holding and persisting evaluation results
    """
    def __init__(self, csv_path=None, sort_column_name=None, ascending=True, incremental=False):
        """
        :param csv_path: path to save the data frame to upon every update
        :param sort_column_name: the column name by which to sort the data frame that is collected; if None, do not sort
        :param ascending: whether to sort in ascending order; has an effect only if sortColumnName is not None
        :param incremental: whether to add to an existing CSV file instead of overwriting it
        """
        self.sort_column_name = sort_column_name
        self.csv_path = csv_path
        self.ascending = ascending
        csv_path_exists = csv_path is not None and os.path.exists(csv_path)
        if csv_path_exists and incremental:
            self.df = pd.read_csv(csv_path)
            log.info(f"Found existing CSV file with {len(self.df)} entries; {csv_path} will be extended (incremental mode)")
            self._current_row = len(self.df)
            self.value_dicts = [nt._asdict() for nt in self.df.itertuples()]
        else:
            if csv_path is not None:
                if not csv_path_exists:
                    log.info(f"Results will be written to new file {csv_path}")
                else:
                    log.warning(f"Results in existing file ({csv_path}) will be overwritten (non-incremental mode)")
            self.df = None
            self._current_row = 0
            self.value_dicts = []

    def add_values(self, values: Dict[str, Any]):
        """
        Adds the provided values as a new row to the collection.
        If csvPath was provided in the constructor, saves the updated collection to that file.

        :param values: Dict holding the evaluation results and parameters
        :return:
        """
        if self.df is None:
            cols = list(values.keys())

            # check sort column and move it to the front
            if self.sort_column_name is not None:
                if self.sort_column_name not in cols:
                    log.warning(f"Specified sort column '{self.sort_column_name}' not in list of columns: {cols}; "
                                f"sorting will not take place!")
                else:
                    cols.remove(self.sort_column_name)
                    cols.insert(0, self.sort_column_name)

            self.df = pd.DataFrame(columns=cols)
        else:
            # check for new columns
            for col in values.keys():
                if col not in self.df.columns:
                    self.df[col] = None

        # append data to data frame
        self.df.loc[self._current_row] = [values.get(c) for c in self.df.columns]
        self._current_row += 1

        # sort where applicable
        if self.sort_column_name is not None and self.sort_column_name in self.df.columns:
            self.df.sort_values(self.sort_column_name, axis=0, inplace=True, ascending=self.ascending)
            self.df.reset_index(drop=True, inplace=True)

        self._save_csv()

    def _save_csv(self):
        if self.csv_path is not None:
            dirname = os.path.dirname(self.csv_path)
            if dirname != "":
                os.makedirs(dirname, exist_ok=True)
            self.df.to_csv(self.csv_path, index=False)

    def get_data_frame(self) -> pd.DataFrame:
        return self.df

    def contains(self, values: Dict[str, Any]):
        for existingValues in self.value_dicts:
            is_contained = True
            for k, v in values.items():
                ev = existingValues.get(k)
                if ev != v and str(ev) != str(v):
                    is_contained = False
                    break
            if is_contained:
                return True


class GridSearch(TrackingMixin):
    """
    Instances of this class can be used for evaluating models with different user-provided parametrizations
    over the same data and persisting the results
    """
    log = log.getChild(__qualname__)

    def __init__(self,
            model_factory: Callable[..., VectorModel],
            parameter_options: Union[Dict[str, Sequence[Any]], List[Dict[str, Sequence[Any]]]],
            num_processes=1,
            csv_results_path: str = None,
            incremental=False,
            incremental_skip_existing=False,
            parameter_combination_skip_decider: ParameterCombinationSkipDecider = None,
            model_save_directory: str = None,
            name: str = None):
        """
        :param model_factory: the function to call with keyword arguments reflecting the parameters to try in order to obtain a model
            instance
        :param parameter_options: a dictionary which maps from parameter names to lists of possible values - or a list of such dictionaries,
            where each dictionary in the list has the same keys
        :param num_processes: the number of parallel processes to use for the search (use 1 to run without multi-processing)
        :param csv_results_path: the path to a directory or concrete CSV file to which the results shall be written;
            if it is None, no CSV data will be written; if it is a directory, a file name starting with this grid search's name (see below)
            will be created.
            The resulting CSV data will contain one line per evaluated parameter combination.
        :param incremental: whether to add to an existing CSV file instead of overwriting it
        :param incremental_skip_existing: if incremental mode is on, whether to skip any parameter combinations that are already present
            in the CSV file
        :param parameter_combination_skip_decider: an instance to which parameters combinations can be passed in order to decide whether the
            combination shall be skipped (e.g. because it is redundant/equivalent to another combination or inadmissible)
        :param model_save_directory: the directory where the serialized models shall be saved; if None, models are not saved
        :param name: the name of this grid search, which will, in particular, be prepended to all saved model files;
            if None, a default name will be generated of the form "gridSearch_<timestamp>"
        """
        self.model_factory = model_factory
        if type(parameter_options) == list:
            self.parameter_options_list = parameter_options
        else:
            self.parameter_options_list = [parameter_options]
        self.param_names = set(self.parameter_options_list[0].keys())
        for d in self.parameter_options_list[1:]:
            if set(d.keys()) != self.param_names:
                raise ValueError("Keys must be the same for all parameter options dictionaries")
        self.num_processes = num_processes
        self.parameter_combination_skip_decider = parameter_combination_skip_decider
        self.model_save_directory = model_save_directory
        self.name = name if name is not None else "gridSearch_" + datetime.now().strftime('%Y%m%d-%H%M%S')
        self.csv_results_path = csv_results_path
        self.incremental = incremental
        self.incremental_skip_existing = incremental_skip_existing
        if self.csv_results_path is not None and os.path.isdir(csv_results_path):
            self.csv_results_path = os.path.join(self.csv_results_path, f"{self.name}_results.csv")

        self.num_combinations = 0
        for parameter_options in self.parameter_options_list:
            n = 1
            for options in parameter_options.values():
                n *= len(options)
            self.num_combinations += n
        log.info(f"Created GridSearch object for {self.num_combinations} parameter combinations")

        self._executor = None

    @classmethod
    def _eval_params(cls,
            model_factory: Callable[..., VectorModel],
            metrics_evaluator: MetricsDictProvider,
            skip_decider: ParameterCombinationSkipDecider,
            grid_search_name, combination_idx,
            model_save_directory: Optional[str],
            **params) -> Optional[Dict[str, Any]]:
        if skip_decider is not None:
            if skip_decider.is_skipped(params):
                cls.log.info(f"Parameter combination is skipped according to {skip_decider}: {params}")
                return None
        cls.log.info(f"Evaluating {params}")
        model = model_factory(**params)
        values = metrics_evaluator.compute_metrics(model)
        if model_save_directory is not None:
            filename = f"{grid_search_name}_{combination_idx}.pickle"
            log.info(f"Saving trained model to {filename} ...")
            model.save(os.path.join(model_save_directory, filename))
            values["filename"] = filename
        values["str(model)"] = str(model)
        values.update(**params)
        if skip_decider is not None:
            skip_decider.tell(params, values)
        return values

    def run(self, metrics_evaluator: MetricsDictProvider, sort_column_name=None, ascending=True) -> "GridSearch.Result":
        """
        Run the grid search. If csvResultsPath was provided in the constructor, each evaluation result will be saved
        to that file directly after being computed

        :param metrics_evaluator: the evaluator or cross-validator with which to evaluate models
        :param sort_column_name: the name of the metric (column) by which to sort the data frame of results; if None, do not sort.
            Note that all Metric instances have a static member `name`, e.g. you could use `RegressionMetricMSE.name`.
        :param ascending: whether to sort in ascending order; has an effect only if `sort_column_name` is specified.
            The result object will assume, by default, that the resulting top/first element is the best,
            i.e. ascending=False means "higher is better", and ascending=True means "Lower is better".
        :return: an object holding the results
        """
        if self.tracked_experiment is not None:
            logging_callback = self.tracked_experiment.track_values
        elif metrics_evaluator.tracked_experiment is not None:
            logging_callback = metrics_evaluator.tracked_experiment.track_values
        else:
            logging_callback = None
        params_metrics_collection = ParametersMetricsCollection(csv_path=self.csv_results_path, sort_column_name=sort_column_name,
            ascending=ascending, incremental=self.incremental)

        def collect_result(values):
            if values is None:
                return
            if logging_callback is not None:
                logging_callback(values)
            params_metrics_collection.add_values(values)
            log.info(f"Updated grid search result:\n{params_metrics_collection.get_data_frame().to_string()}")

        if self.num_processes == 1:
            combination_idx = 0
            for parameter_options in self.parameter_options_list:
                for params_dict in iter_param_combinations(parameter_options):
                    if self.incremental_skip_existing and self.incremental:
                        if params_metrics_collection.contains(params_dict):
                            log.info(f"Skipped because parameters are already present in collection (incremental mode): {params_dict}")
                            continue
                    collect_result(self._eval_params(self.model_factory, metrics_evaluator, self.parameter_combination_skip_decider,
                        self.name, combination_idx, self.model_save_directory, **params_dict))
                    combination_idx += 1
        else:
            executor = ProcessPoolExecutor(max_workers=self.num_processes)
            futures = []
            combination_idx = 0
            for parameter_options in self.parameter_options_list:
                for params_dict in iter_param_combinations(parameter_options):
                    if self.incremental_skip_existing and self.incremental:
                        if params_metrics_collection.contains(params_dict):
                            log.info(f"Skipped because parameters are already present in collection (incremental mode): {params_dict}")
                            continue
                    futures.append(executor.submit(self._eval_params, self.model_factory, metrics_evaluator,
                        self.parameter_combination_skip_decider,
                        self.name, combination_idx, self.model_save_directory, **params_dict))
                    combination_idx += 1
            for future in futures:
                collect_result(future.result())

        df = params_metrics_collection.get_data_frame()
        return self.Result(df, self.param_names, default_metric_name=sort_column_name, default_higher_is_better=not ascending)

    class Result:
        def __init__(self, df: pd.DataFrame, param_names: List[str], default_metric_name: Optional[str] = None,
                default_higher_is_better: Optional[bool] = None):
            self.df = df
            self.param_names = param_names
            self.default_metric_name = default_metric_name
            if default_metric_name is not None:
                self.default_higher_is_better = default_higher_is_better
            else:
                self.default_higher_is_better = None

        @dataclass
        class BestParams:
            metric_name: str
            metric_value: float
            params: dict

        def get_best_params(self, metric_name: Optional[str] = None, higher_is_better: Optional[bool] = None) -> BestParams:
            """

            :param metric_name: the metric name for which to return the best result; can be None if the GridSearch used
                a metric to sort by
            :param higher_is_better: whether higher is better for the metric to sort by; can be None if the GridSearch
                use a metric to sort by and configured the sort order such that the best configuration is at the top
            :return: a pair (d, v) where d dictionary with the best parameters found during the grid search and v is the
                corresponding metric value
            """
            if metric_name is None:
                metric_name = self.default_metric_name
                if metric_name is None:
                    raise ValueError("metric_name must be specified")
            if higher_is_better is None:
                higher_is_better = self.default_higher_is_better
                if higher_is_better is None:
                    raise ValueError("higher_is_better must be specified")

            df = self.df.sort_values(metric_name, axis=0, inplace=False, ascending=not higher_is_better)

            best_params = {}
            for param in self.param_names:
                best_params[param] = df.iloc[0][param]
            best_metric_value = df.iloc[0][metric_name]

            return self.BestParams(metric_name=metric_name, metric_value=best_metric_value, params=best_params)


class SAHyperOpt(TrackingMixin):
    log = log.getChild(__qualname__)

    class State(SAState):
        def __init__(self, params: Dict[str, Any], random_state: Random, results: Dict, compute_metric: Callable[[Dict[str, Any]], float]):
            self.compute_metric = compute_metric
            self.results = results
            self.params = dict(params)
            super().__init__(random_state)

        def compute_cost_value(self) -> SACostValueNumeric:
            return SACostValueNumeric(self.compute_metric(self.params))

        def get_state_representation(self):
            return self.params

        def apply_state_representation(self, representation):
            self.results.update(representation)

    class ParameterChangeOperator(SAOperator[State]):
        def __init__(self, state: 'SAHyperOpt.State'):
            super().__init__(state)

        def apply_state_change(self, params):
            self.state.params.update(params)

        def cost_delta(self, params) -> SACostValue:
            model_params = dict(self.state.params)
            model_params.update(params)
            return SACostValueNumeric(self.state.compute_metric(model_params) - self.state.cost.value())

        def choose_params(self) -> Optional[Tuple[Tuple, Optional[SACostValue]]]:
            params = self._choose_changed_model_parameters()
            if params is None:
                return None
            return ((params, ), None)

        @abstractmethod
        def _choose_changed_model_parameters(self) -> Dict[str, Any]:
            pass

    def __init__(self,
            model_factory: Callable[..., VectorModel],
            ops_and_weights: List[Tuple[Callable[['SAHyperOpt.State'], 'SAHyperOpt.ParameterChangeOperator'], float]],
            initial_parameters: Dict[str, Any],
            metrics_evaluator: MetricsDictProvider,
            metric_to_optimise: str,
            minimise_metric: bool = False,
            collect_data_frame: bool = True,
            csv_results_path: Optional[str] = None,
            parameter_combination_equivalence_class_value_cache: ParameterCombinationEquivalenceClassValueCache = None,
            p0: float = 0.5,
            p1: float = 0.0):
        """
        :param model_factory: a factory for the generation of models which is called with the current parameter combination
            (all keyword arguments), initially initialParameters
        :param ops_and_weights: a sequence of tuples (operator factory, operator weight) for simulated annealing
        :param initial_parameters: the initial parameter combination
        :param metrics_evaluator: the evaluator/validator to use in order to evaluate models
        :param metric_to_optimise: the name of the metric (as generated by the evaluator/validator) to optimise
        :param minimise_metric: whether the metric is to be minimised; if False, maximise the metric
        :param collect_data_frame: whether to collect (and regularly log) the data frame of all parameter combinations and
            evaluation results
        :param csv_results_path: the (optional) path of a CSV file in which to store a table of all computed results;
            if this is not None, then collectDataFrame is automatically set to True
        :param parameter_combination_equivalence_class_value_cache: a cache in which to store computed results and whose notion
            of equivalence can be used to avoid duplicate computations
        :param p0: the initial probability (at the start of the optimisation) of accepting a state with an inferior evaluation
            to the current state's (for the mean observed evaluation delta)
        :param p1: the final probability (at the end of the optimisation) of accepting a state with an inferior evaluation
            to the current state's (for the mean observed evaluation delta)
        """
        self.minimise_metric = minimise_metric
        self.evaluator_or_validator = metrics_evaluator
        self.metric_to_optimise = metric_to_optimise
        self.initial_parameters = initial_parameters
        self.ops_and_weights = ops_and_weights
        self.model_factory = model_factory
        self.csv_results_path = csv_results_path
        if csv_results_path is not None:
            collect_data_frame = True
        self.parameters_metrics_collection = ParametersMetricsCollection(csv_path=csv_results_path) if collect_data_frame else None
        self.parameter_combination_equivalence_class_value_cache = parameter_combination_equivalence_class_value_cache
        self.p0 = p0
        self.p1 = p1
        self._sa = None

    @classmethod
    def _eval_params(cls,
            model_factory,
            metrics_evaluator: MetricsDictProvider,
            parameters_metrics_collection: Optional[ParametersMetricsCollection],
            parameter_combination_equivalence_class_value_cache,
            tracked_experiment: Optional[TrackedExperiment],
            **params):
        if tracked_experiment is not None and metrics_evaluator.tracked_experiment is not None:
            log.warning(f"Tracked experiment already set in evaluator, results will be tracked twice and"
                        f"might get overwritten!")

        metrics = None
        if parameter_combination_equivalence_class_value_cache is not None:
            metrics = parameter_combination_equivalence_class_value_cache.get(params)
        if metrics is not None:
            cls.log.info(f"Result for parameter combination {params} could be retrieved from cache, not adding new result")
        else:
            cls.log.info(f"Evaluating parameter combination {params}")
            model = model_factory(**params)
            metrics = metrics_evaluator.compute_metrics(model)
            cls.log.info(f"Got metrics {metrics} for {params}")

            values = dict(metrics)
            values["str(model)"] = str(model)
            values.update(**params)
            if tracked_experiment is not None:
                tracked_experiment.track_values(values)
            if parameters_metrics_collection is not None:
                parameters_metrics_collection.add_values(values)
                cls.log.info(f"Data frame with all results:\n\n{parameters_metrics_collection.get_data_frame().to_string()}\n")
            if parameter_combination_equivalence_class_value_cache is not None:
                parameter_combination_equivalence_class_value_cache.set(params, metrics)
        return metrics

    def _compute_metric(self, params: Dict[str, Any]):
        metrics = self._eval_params(self.model_factory, self.evaluator_or_validator, self.parameters_metrics_collection,
            self.parameter_combination_equivalence_class_value_cache, self.tracked_experiment, **params)
        metric_value = metrics[self.metric_to_optimise]
        if not self.minimise_metric:
            return -metric_value
        return metric_value

    def run(self, max_steps: Optional[int] = None, duration: Optional[float] = None, random_seed: int = 42, collect_stats: bool = True):
        sa = SimulatedAnnealing(lambda: SAProbabilitySchedule(None, SAProbabilityFunctionLinear(p0=self.p0, p1=self.p1)),
            self.ops_and_weights, max_steps=max_steps, duration=duration, random_seed=random_seed, collect_stats=collect_stats)
        results = {}
        self._sa = sa
        sa.optimise(lambda r: self.State(self.initial_parameters, r, results, self._compute_metric))
        return results

    def get_simulated_annealing(self) -> SimulatedAnnealing:
        return self._sa
